import logging

from twisted.internet import reactor

from trosnoth.const import TICK_PERIOD, INITIAL_ASSUMED_LATENCY
from trosnoth.messages import (
    JoinRequestMsg, TickMsg, ResyncPlayerMsg, UpdatePlayerStateMsg,
    AimPlayerAtMsg, UpgradeApprovedMsg, PlayerHasUpgradeMsg, ShootMsg,
    CheckSyncMsg, WorldResetMsg, BuyUpgradeMsg, GameOverMsg,
)
from trosnoth.model.shot import LocalShot, LocalGrenade
from trosnoth.model.upgrades import Grenade
from trosnoth.utils.event import Event
from trosnoth.utils.message import MessageConsumer

log = logging.getLogger(__name__)

SYNC_CHECK_PERIOD = 3 / TICK_PERIOD


class Agent(object):
    '''
    Base class for things which can be connected to a Game using Game.addAgent.
    This may represent a user interface, an AI player, someone connecting over
    the network, or anything else that wants to receive interact with the game.
    '''

    def __init__(self, game, *args, **kwargs):
        super(Agent, self).__init__(*args, **kwargs)
        self.game = game
        self.player = None
        self.game.onServerCommand.addListener(self.gotServerCommand)

    def stop(self):
        '''
        Disconnects this agent from things that it's subscribed to and stops
        any active timed or looping calls.
        '''
        self.game.onServerCommand.removeListener(self.gotServerCommand)

    def gotServerCommand(self, msg):
        raise NotImplementedError(
            '%s.gotServerCommand' % (self.__class__.__name__,))

    def setPlayer(self, player):
        '''
        Called by the connected Game object when we are given authority to
        control the specified player. Also called with player=None when we no
        longer have the authority to control any player.
        '''
        self.player = player

    def messageToAgent(self, msg):
        '''
        Called by the connected Game object when there is a message
        specifically for this Agent (as opposed to a general game message).
        '''
        raise NotImplementedError(
            '%s.messageToAgent' % (self.__class__.__name__,))


class ConcreteAgent(Agent, MessageConsumer):
    '''
    Base class for Agents that actually represent the world (rather than
    proxying it through to the Network or some other agent).

    Note that a concrete agent is designed to represent only zero or one
    players.
    '''

    def __init__(self, *args, **kwargs):
        super(ConcreteAgent, self).__init__(*args, **kwargs)
        self.world = self.game.world
        self.localState = LocalState(self)
        self.lastPlayerAimSent = (None, None)
        self.nextSyncCheck = None

    def gotServerCommand(self, msg):
        msg.tracePoint(self, 'gotServerCommand')
        self.consumeMsg(msg)

    def messageToAgent(self, msg):
        msg.tracePoint(self, 'messageToAgent')
        self.consumeMsg(msg)

    def sendRequest(self, msg):
        msg.tracePoint(self, 'sendRequest')
        if not msg.clientValidate(
                self.localState, self.world, self._validationResponse):
            return
        msg.applyRequestToLocalState(self.localState)
        if isinstance(msg, AimPlayerAtMsg):
            # Special case: apply aim messages to the local state, but don't
            # indiscriminately send them to the server as they are mostly
            # cosmetic.
            return
        if isinstance(msg, ShootMsg) or (isinstance(msg, BuyUpgradeMsg)
                and msg.upgradeType == 'g'):
            # Special case: before shooting make sure the server knows what
            # direction we are facing.
            self.maybeSendAimMsg(self.world.lastTickId)
        reactor.callLater(0, self.game.agentRequest, self, msg)

    def _validationResponse(self, msg):
        self.consumeMsg(msg)

    def defaultHandler(self, msg):
        msg.tracePoint(self, 'defaultHandler')
        msg.applyOrderToLocalState(self.localState, self.world)

    def setPlayer(self, player):
        super(ConcreteAgent, self).setPlayer(player)
        if player:
            self.localState.playerJoined(player)
            self.resyncLocalPlayer({})
        else:
            self.localState.lostPlayer()

    @GameOverMsg.handler
    def handle_GameOverMsg(self, msg):
        if self.player is not None:
            self.localState.player.readyToStart = False
            self.localState.player.readyForTournament = False

    @WorldResetMsg.handler
    def handle_WorldResetMsg(self, msg):
        if self.player is not None:
            oldKeyState = dict(self.localState.player._state)
            self.localState.refreshPlayer()
            self.resyncLocalPlayer(oldKeyState)

    @TickMsg.handler
    def handle_TickMsg(self, msg):
        self.maybeSendAimMsg(tickId=max(0, msg.tickId - 1))
        self.localState.tick()

        player = self.localState.player
        if player and self.nextSyncCheck <= msg.tickId:
            self.nextSyncCheck = msg.tickId + SYNC_CHECK_PERIOD
            self.sendRequest(CheckSyncMsg(
                self.world.lastTickId, player.pos[0],
                player.pos[1], player.yVel))

    def maybeSendAimMsg(self, tickId):
        '''
        Tell the server where the local player is facing.
        '''
        player = self.localState.player
        if player:
            lastAngle, lastThrust = self.lastPlayerAimSent
            if (
                    lastAngle != player.angleFacing or
                    lastThrust != player.ghostThrust):
                reactor.callLater(
                    0, self.game.agentRequest, self, AimPlayerAtMsg(
                        player.angleFacing, player.ghostThrust, tickId))
                self.lastPlayerAimSent = (
                    player.angleFacing, player.ghostThrust)

    @ResyncPlayerMsg.handler
    def handle_ResyncPlayerMsg(self, msg):
        oldKeyState = dict(self.localState.player._state)
        self.localState.player.applyPlayerUpdate(msg)
        self.resyncLocalPlayer(oldKeyState)

    def resyncLocalPlayer(self, oldKeyState):
        newKeyState = self.localState.player._state
        self.sendRequest(self.localState.player.buildResyncAcknowledgement())
        self.nextSyncCheck = self.world.lastTickId + SYNC_CHECK_PERIOD

        for key, value in oldKeyState.iteritems():
            if value != newKeyState[key]:
                self.sendRequest(UpdatePlayerStateMsg(
                    value, self.world.lastTickId, stateKey=key))

    @UpgradeApprovedMsg.handler
    def handle_UpgradeApprovedMsg(self, msg):
        self.sendRequest(PlayerHasUpgradeMsg(
            msg.upgradeType, self.world.lastTickId))

    def sendJoinRequest(self, teamId, nick, authTag=0, bot=False):
        if self.player is not None:
            raise RuntimeError('Already joined.')

        msg = JoinRequestMsg(
            teamId, authTag=authTag, nick=nick.encode(), bot=bot)
        if bot:
            msg.localBotRequest = True

        self.sendRequest(msg)


class LocalState(object):
    '''
    Stores state information which a client wants to keep which do not need to
    wait for a round trip to the server. e.g. when a player moves, it should
    start moving on the local screen even before the server has received the
    message.
    '''

    LOCAL_ID_CAP = 1 << 16

    def __init__(self, agent):
        self.agent = agent
        self.world = agent.world
        self.player = None
        self.onShoxwave = Event()
        self.shotById = {}
        self.localShots = {}
        self.localGrenade = None
        self.nextLocalId = 1
        self.serverDelay = INITIAL_ASSUMED_LATENCY
        self.world.onShotRemoved.addListener(self.shotRemoved)

    @property
    def shots(self):
        for shot in self.shotById.itervalues():
            yield shot
        for shot in self.localShots.itervalues():
            yield shot

    def playerJoined(self, player):
        # Create a shallow copy of the player object, and use it to simulate
        # movement before the server approves it.
        self.player = player.clone()

        player.onZombieHit.addListener(self.noticedZombieHit)
        player.onStarsChanged.addListener(self.playerStarsChanged)
        player.onGotTrosball.addListener(self.gotTrosball)
        player.onDied.addListener(self.died)

    def refreshPlayer(self):
        realPlayer = self.world.playerWithId[self.player.id]
        self.player.restore(realPlayer.dump())

    def noticedZombieHit(self, killer, shot, deathType):
        self.player.properHit(killer, deathType)

    def playerStarsChanged(self):
        if self.player.dead:
            return
        player = self.world.getPlayer(self.player.id)
        self.player.stars = player.stars

    def gotTrosball(self):
        self.player.deleteUpgrade()

    def died(self, killer, deathType):
        self.localShots.clear()

    def lostPlayer(self):
        self.player.onZombieHit.removeListener(self.noticedZombieHit)
        self.player.onStarsChanged.removeListener(self.playerStarsChanged)
        self.player.onGotTrosball.removeListener(self.gotTrosball)
        self.player.onDied.removeListener(self.died)
        self.player = None

    def shotFired(self):
        if self.player.shoxwave:
            self.onShoxwave()
            localId = 0
        else:
            localId = self.nextLocalId
            self.nextLocalId = (self.nextLocalId + 1) % self.LOCAL_ID_CAP
            self.localShots[localId] = self.player.createShot(
                shotClass=LocalShot)
        self.player.weaponDischarged()
        return localId

    def matchShot(self, localId, shotId):
        if localId in self.localShots:
            shot = self.localShots.pop(localId)
            shot.realShotStarted = True
            self.shotById[shotId] = shot

    def shotRemoved(self, shotId, *args, **kwargs):
        if shotId in self.shotById:
            del self.shotById[shotId]

    def grenadeLaunched(self):
        self.localGrenade = LocalGrenade(
            self, self.world, self.player, Grenade.timeRemaining)

    def matchGrenade(self):
        for grenade in self.world.grenades:
            if grenade.player.id == self.player.id:
                self.localGrenade.realShotStarted = True
                return

    def grenadeRemoved(self):
        self.localGrenade = None

    def tick(self):
        if self.player:
            self.player.reset()
            self.player.advance()

            if not self.player.dead:
                for unit in self.world.getCollectableUnits():
                    if unit.hitLocalPlayer:
                        continue
                    if unit.checkCollision(self.player, 0):
                        unit.collidedWithLocalPlayer(self.player)

        for shot in self.shotById.values() + self.localShots.values():
            shot.reset()
            shot.advance()
        if self.localGrenade:
            self.localGrenade.reset()
            self.localGrenade.advance()

        for shotId, shot in self.shotById.items():
            if shot.expired:
                del self.shotById[shotId]
        for localId, shot in self.localShots.items():
            if shot.expired:
                del self.localShots[localId]



import logging

from trosnoth.const import (
    GAME_FULL_REASON, UNAUTHORISED_REASON, NICK_USED_REASON, BAD_NICK_REASON,
    USER_IN_GAME_REASON,
)
from trosnoth.trosnothgui.ingame.dialogs import (
    JoinGameDialog, JoiningDialog, SpectateResult,
)
from trosnoth.gui.framework.dialogbox import DialogResult
from trosnoth.model import gameStates

log = logging.getLogger(__name__)


class JoinGameController(object):
    '''
    Shows and hides the join game dialog.
    '''

    def __init__(self, app, gameInterface, game):
        self.app = app
        self.world = game.world
        self.interface = gameInterface
        self.joined = False
        self.joiningInfo = None
        self.gameInterface = gameInterface

        self.joiningScreen = JoiningDialog(self.app, self)

        self.joinGameDialog = JoinGameDialog(self.app, self, self.world)
        self.joinGameDialog.onClose.addListener(self._joinDlgClose)

    def start(self):
        '''
        Called by GameInterface when the player's agent has been added.
        '''
        self.maybeShowJoinDialog(autoJoin=True)

    def maybeShowJoinDialog(self, autoJoin=False):
        if self.world.isIncomplete:
            # Have not received full world data. Wait until reset message is
            # received.
            return

        if self.interface.player:
            return

        nick = self.app.identitySettings.nick
        if autoJoin and nick and (self.world.isServer or self.world.state in (
                gameStates.Lobby, gameStates.Ended)):
            # Practise or local game. Join automatically.
            self.attemptGameJoin(nick, None)
        elif not self.joinGameDialog.showing:
            self.joinGameDialog.show()

    def gotWorldReset(self):
        '''
        Called by GameInterface whenever a world reset is received.
        '''
        self.maybeShowJoinDialog()

    def hide(self):
        if self.joinGameDialog.showing:
            self.joinGameDialog.close()
        if self.joiningScreen.showing:
            self.joiningScreen.close()

    def _joinDlgClose(self):
        if self.joinGameDialog.result is None:
            return
        if self.joinGameDialog.result == SpectateResult:
            self.interface.spectate()
        elif self.joinGameDialog.result != DialogResult.OK:
            self.interface.joinDialogCancelled()
        else:
            nick = self.joinGameDialog.nickBox.value.strip()
            self.app.identitySettings.setNick(nick)

            team = self.joinGameDialog.selectedTeam
            self.attemptGameJoin(nick, team)

    def attemptGameJoin(self, nick, team):
        self.joiningInfo = nick, team
        self.joiningScreen.show(nick)
        self.interface.joinGame(nick, team)

    def showMessage(self, text):
        self.joinGameDialog.cantJoinYet.setText(text)

    def cancelJoin(self, sender):
        # TODO: This doesn't actually cancel anything
        self.joiningScreen.close()
        self.maybeShowJoinDialog()

    def joinFailed(self, reason):
        if self.joiningScreen.showing:
            self.joiningScreen.close()
        self.maybeShowJoinDialog()

        if reason == GAME_FULL_REASON:
            # Team is full.
            self.showMessage('That team is full!')
        elif reason == NICK_USED_REASON:
            # Nickname is already being used.
            self.showMessage('That name is already being used!')
        elif reason == BAD_NICK_REASON:
            # Nickname is too short or long.
            self.showMessage('That name is not allowed!')
        elif reason == UNAUTHORISED_REASON:
            self.showMessage('You are not authorised to join!')
        elif reason == USER_IN_GAME_REASON:
            self.showMessage('Cannot join the same game twice!')
        elif reason == 'timeout':
            self.showMessage('Join timed out.')
        else:
            # Unknown reason.
            self.showMessage('Join failed: %r' % (reason,))



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.app import Application
from trosnoth.gui.framework import framework, elements, scrollableCanvas
from trosnoth.gui.common import *
from trosnoth.gui.fonts.font import Font
import pygame

class Interface(framework.CompoundElement):
    def __init__(self, app):
        super(Interface, self).__init__(app)
        backdrop = pygame.image.load('Awesome Leaders.jpg').convert()
        x, y = backdrop.get_size()
        size = Size(x, y)
        sc1 = scrollableCanvas.ScrollableCanvas(app, ScaledLocation(20,20), size, ScaledSize(200,200))
        sc2 = scrollableCanvas.ScrollableCanvas(app, ScaledLocation(230,20), size, ScaledSize(200,y+scrollableCanvas.ScrollBar.defaultWidth))
        sc3 = scrollableCanvas.ScrollableCanvas(app, ScaledLocation(440,20), size, ScaledSize(x,y))
        sc4 = scrollableCanvas.ScrollableCanvas(app, ScaledLocation(20,530), size, ScaledSize(x+scrollableCanvas.ScrollBar.defaultWidth, 200))
        #sc5 = scrollableCanvas.ScrollableCanvas(app, ScaledLocation(20,530), size, ScaledSize(x+scrollableCanvas.ScrollBar.defaultWidth, y))
        self.elements = [sc1, sc2, sc3, sc4]#, sc5]
        font = Font("KLEPTOCR.TTF", 30)
        onclick = lambda sender: self.app.screenManager.setScreenProperties((1680, 1050), 0, "Testing")
        for sc in self.elements:
            sc.elements = [elements.PictureElement(app, backdrop, Location(scrollableCanvas.ScrollableCanvasAttachedPoint(sc, (0,0)), 'topleft')),
                           elements.TextButton(app, Location(scrollableCanvas.ScrollableCanvasAttachedPoint(sc, (20,20))), "Hello", font, (0,128,0), (0,0,128), onClick = onclick)]

    def draw(self, screen):
        screen.fill((0,0,0))
        super(Interface, self).draw(screen)

size = (1200,768)

if __name__ == "__main__":
    a = Application(size, 0, "Testing", Interface)
    a.run()



from math import pi, atan2
import random

from trosnoth.utils.math import distance
from trosnoth.ai import AI
from trosnoth.utils.twist import WeakCallLater


class AlphaAI(AI):
    nick = 'AlphaBot'
    playable = True

    def start(self):
        self.xdir = 'left'
        self.dead = True

        self.clRespawn = None
        self.clChangeXDir = None
        self.clJump = None
        self.clDrop = None
        self.clShoot = None

        self._launchDeadCallLaters(0)

    def disable(self):
        for cl in (self.clRespawn, self.clChangeXDir, self.clJump, self.clDrop,
                self.clShoot):
            if cl:
                cl.cancel()

    def _launchAliveCallLaters(self):
        self._cancelAliveCallLaters()
        self.clChangeXDir = WeakCallLater(0, self, 'changeXDir')
        self.clJump = WeakCallLater(0, self, 'jumpAgain')
        self.clDrop = WeakCallLater(0, self, 'drop')
        self.clShoot = WeakCallLater(0, self, 'shoot')

    def _cancelAliveCallLaters(self):
        if self.clChangeXDir is not None and self.clChangeXDir.active():
            self.clChangeXDir.cancel()
            self.clChangeXDir = None
        if self.clJump is not None and self.clJump.active():
            self.clJump.cancel()
            self.clJump = None
        if self.clDrop is not None and self.clDrop.active():
            self.clDrop.cancel()
            self.clDrop = None
        if self.clShoot is not None and self.clShoot.active():
            self.clShoot.cancel()
            self.clShoot = None

    def _launchDeadCallLaters(self, respawnTime=None):
        if respawnTime == None:
            respawnTime = self.player.respawnGauge
        self._cancelDeadCallLaters()
        self.clRespawn = WeakCallLater(respawnTime, self, 'tryRespawn')

    def _cancelDeadCallLaters(self):
        if self.clRespawn is not None:
            self.clRespawn.cancel()
            self.clRespawn = None

    def died(self, killer, deathType):
        self.dead = True
        self._cancelAliveCallLaters()
        self._launchDeadCallLaters()
        self.aimAtActionZone()

    def respawned(self):
        self.dead = False
        self._cancelDeadCallLaters()
        self._launchAliveCallLaters()

    def aimAtNearestFriendlyZone(self, time = None):
        if time == None:
            time = self.player.respawnGauge
        nearest = None
        for zone in self.world.zones:
            if zone.owner == self.player.team:
                if nearest == None or (distance(zone.defn.pos, self.player.pos)
                        < distance(self.player.pos, nearest.defn.pos)):
                    nearest = zone
        if nearest == None:
            # Stop moving
            self.aimAtPoint(self.player.pos, time)
        else:
            self.aimAtPoint(nearest.defn.pos, time)

    def aimAtActionZone(self, time = None):
        if time == None:
            time = self.player.respawnGauge
        zone = self.getBestRespawnZone()

        self.aimAtPoint(zone.defn.pos, time)

    def getBestRespawnZone(self):
        best, score = None, -1000
        for zone in self.world.zones:
            thisScore = self.getRespawnScore(zone)

            if thisScore > score:
                best, score = zone, thisScore
        return best

    def getRespawnScore(self, zone):
        thisScore = 0
        threat = False
        team = self.player.team
        if zone.owner == team:
            for adjDef, open in zone.defn.zones_AdjInfo():
                adj = self.world.zoneWithDef[adjDef]
                # Find boundary zone.
                if adj.isNeutral():
                    if open:
                        thisScore += 3
                elif adj.owner != team:
                    threat = True
                    if open:
                        thisScore += 5
                if open:
                    # Add 2 points for each unmarked enemy in an open, adjacent
                    # zone
                    thisScore += self.numUnmarkedEnemies(adj) * 2
            if threat == False:
                # No threat, since no enemy adjacent zones
                return thisScore
            else:
                thisScore += 5

            # And 20 points for each unmarked enemy in this zone
            thisScore += self.numUnmarkedEnemies(zone) * 20
        return thisScore

    def numUnmarkedEnemies(self, zone):
        score = 0
        for player in zone.players:
            if player.dead:
                continue
            if player.team == self.player.team:
                score -= 1
            else:
                score += 1
        return max(score, 0)

    def aimAtNearestEnemy(self):
        nearest = None
        thisPlayer = self.player
        for player in self.world.players:
            if (not player.isFriendsWith(thisPlayer) and not player.dead and
                    not player.turret and not player.invisible):
                if nearest == None or (distance(player.pos, thisPlayer.pos) <
                        distance(thisPlayer.pos, nearest.pos)):
                    nearest = player
        if nearest is not None:
            # Add uncertainty
            pos = (nearest.pos[0] + (random.random() * 80 - 40),
                   nearest.pos[1] + (random.random() * 80 - 40))
            self.aimAtPoint(pos)

    def aimAtPoint(self, pos, timeLeft = None):
        dx = pos[0] - self.player.pos[0]
        dy = pos[1] - self.player.pos[1]
        theta = atan2(dx, -dy)
        thrust = 1.0
        if timeLeft is not None:
            dist = (dx ** 2 + dy ** 2) ** 0.5
            maxSpeed = self.world.physics.playerMaxGhostVel
            if timeLeft <= 0:
                speedReq = maxSpeed
            else:
                speedReq = dist / timeLeft

            if speedReq < maxSpeed:
                thrust = speedReq / maxSpeed
        self.doAimAt(theta, thrust)

    def tryRespawn(self):
        if self.dead:
            self.aimAtActionZone()
            self.clRespawn = WeakCallLater(0.5, self, 'tryRespawn')
            self.doRespawn()

    def changeXDir(self):
        t = random.random() * 5 + 3
        self.clChangeXDir = WeakCallLater(t, self, 'changeXDir')
        if self.xdir == 'left':
            self.xdir = 'right'
            self.doAimAt(pi/2.)
            self.doMoveRight()
        else:
            self.xdir = 'left'
            self.doAimAt(-pi/2.)
            self.doMoveLeft()

    def drop(self):
        self.clDrop = WeakCallLater(random.random() * 3 + 1.5, self, 'drop')
        self.doDrop()

    def jumpAgain(self):
        self.clJump = WeakCallLater(random.random() * 0.5 + 0.5, self,
                'stopJump')
        self.doJump()

    def stopJump(self):
        self.clJump = WeakCallLater(random.random() * 1.5, self, 'jumpAgain')
        self.doStopJump()

    def shoot(self):
        self.clShoot = WeakCallLater(random.random() * .5 + .4, self, 'shoot')
        self.aimAtNearestEnemy()
        self.doShoot()

    def zoneTagged(self, zoneId, playerId):
        # In case our zone was tagged
        self.aimAtActionZone()

AIClass = AlphaAI



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

'''
xmpp.py - an XMPP client that can be used to broadcast messages to all its
contacts.

>>> xmppclient = NotificationClient('username@gmail.com', 'password')
>>> xmppclient.startService()
>>> reactor.callLater(3, xmppclient.broadcast, 'Why hello there!')

>>> reactor.run()
'''
import logging

from twisted.words.xish import domish
from wokkel.xmppim import (MessageProtocol, AvailablePresence,
        RosterClientProtocol, PresenceClientProtocol)
from twisted.words.protocols.jabber.jid import internJID
from wokkel.client import XMPPClient
from twisted.internet import reactor

log = logging.getLogger('xmpp')

class NotificationClient(XMPPClient):
    def __init__(self, jidString, password, host=None, port=5222):
        XMPPClient.__init__(self, internJID(jidString), password, host=host,
                port=port)
        self.logTraffic = False

        self.messagebot = MessageBot()
        self.messagebot.setHandlerParent(self)

        self.rosterManager = RosterManager()
        self.rosterManager.setHandlerParent(self)

    @property
    def onlineContacts(self):
        return self.rosterManager.onlineContacts

    def broadcast(self, message):
        for contact in self.onlineContacts:
            try:
                self.sendChat(contact.userhostJID(), message)
            except:
                log.error('Error sending message', exc_info=True)

    def sendChat(self, user, body):
        log.debug('Sending message %r', body)
        msg = domish.Element((None, 'message'))
        msg['to'] = user.full()
        msg['from'] = self.jid.full()
        msg['type'] = 'chat'
        msg.addElement('body', content=body)
        self.send(msg)

    def getNickname(self, jid):
        result = self.rosterManager.nicknames.get(jid.userhostJID())
        if not result:
            result = jid.userhost()
        return result


class RosterManager(RosterClientProtocol, PresenceClientProtocol):
    welcomeMessage = ('Welcome. I will send you notifications of events on '
        'play.trosnoth.org. If you send me messages, I will pass them on '
        'to everyone who is chatting with me. Your email may become visible '
        ' if you do this.')

    onlineContacts = None
    nicknames = None

    maxDelay = 300
    initialDelay = 1.0
    factor = 1.618

    delay = initialDelay
    retries = 0

    def connectionInitialized(self):
        RosterClientProtocol.connectionInitialized(self)
        PresenceClientProtocol.connectionInitialized(self)
        self.attemptGetRoster()

    def attemptGetRoster(self):
        self.getRoster().addCallback(self.gotRoster).addErrback(self.errRoster)

    def gotRoster(self, roster):
        log.debug('got roster')
        self.onlineContacts = set()
        self.nicknames = {}
        for name, item in roster.iteritems():
            jid = item.jid.userhostJID()
            self.nicknames[jid] = item.name or name

            if item.ask:
                self.approveSubscription(item.jid)
            if not item.subscriptionTo:
                self.subscribe(jid)

    def approveSubscription(self, jid):
        self.subscribed(jid)
        self.parent.sendChat(jid, self.welcomeMessage)

    def errRoster(self, reason):
        log.error('Error getting roster:')
        reason.printBriefTraceback()
        self.retry()

    def retry(self):
        self.retries += 1
        self.delay = min(self.delay * self.factor, self.maxDelay)
        reactor.callLater(self.delay, self.attemptGetRoster)

    def onRosterSet(self, item):
        log.debug('roster set: %r', item.jid.userhost())
        self.nicknames[item.jid.userhostJID()] = item.name

    def onRosterRemove(self, entity):
        log.debug('roster remove: %r', entity.userhost())
        if self.onlineContacts is not None:
            self.onlineContacts.discard(entity.userhostJID())

    def subscribeReceived(self, sender):
        '''
        Someone has tried to subscribe to our presence. Approve and subscribe to
        their presence.
        '''
        self.approveSubscription(sender)
        self.subscribe(sender.userhostJID())

    def availableReceived(self, sender, show=None, statuses=None, priority=0):
        '''
        One of our followers has come online.
        '''
        log.debug('online: %r', sender.userhost())
        self.onlineContacts.add(sender.userhostJID())

    def unavailableReceived(self, sender, statuses=None):
        '''
        One of our followers is no longer online.
        '''
        log.debug('offline: %r', sender.userhost())
        self.onlineContacts.discard(sender.userhostJID())

class MessageBot(MessageProtocol):
    def connectionMade(self):
        log.debug('Connected!')

        # send initial presence
        self.send(AvailablePresence())

    def connectionLost(self, reason):
        log.debug('Disconnected!')

    def onMessage(self, msg):
        if msg['type'] == 'chat' and hasattr(msg, 'body'):
            sender = internJID(msg['from'])
            body = msg.body
            self.onChatMessage(sender, body)

    def onChatMessage(self, sender, body):
        sender = sender.userhostJID()
        contents = '%s: %s' % (self.parent.getNickname(sender), body)
        for jid in self.parent.onlineContacts:
            jid = jid.userhostJID()
            if jid != sender:
                self.parent.sendChat(jid, contents)




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.app import Application
from trosnoth.gui.framework import slider, elements
from trosnoth.gui.fonts.font import Font
from trosnoth.gui.common import *
from trosnoth.gui.framework import framework
import pygame

class Interface(framework.CompoundElement):
    def __init__(self, app):
        super(Interface, self).__init__(app)
        backdrop = pygame.Surface(app.screenManager.scaledSize)
        backdrop.fill((128,0,128))
        s1 = slider.Slider(app, ScaledArea(100,100,400,60))
        #s1.setRange(50,150)
        t1 = elements.TextElement(app, "100", Font(None, 60), ScaledLocation(100, 200))
        t2 = elements.TextElement(app, "100", Font(None, 60), ScaledLocation(550, 100))
        onSlide = lambda x: t1.setText("%.1f" % x)
        onChange = lambda x: t2.setText("%.1f" % x)
        s1.onSlide.addListener(onSlide)
        s1.onValueChanged.addListener(onChange)
        self.elements = [elements.PictureElement(app, backdrop), s1, t1, t2]

size = (600,450)

if __name__ == '__main__':
    a = Application(size, 0, "Testing", Interface)
    a.run()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from framework import Element
from trosnoth.utils.event import Event
from trosnoth.gui import keyboard
import pygame

class Hotkey(Element):
    def __init__(self, app, key, modifiers):
        super(Hotkey, self).__init__(app)
        self.onTriggered = Event()
        self.key = key
        self.mods = modifiers

    def processEvent(self, event):
        if event.type == pygame.KEYDOWN:
            if self.testEquals(event.key, event.mod):
                self.onTriggered.execute()
                return
        return event

    def __str__(self):
        return keyboard.shortcutName(self.key, self.mods)

    def testEquals(self, key, modifiers, mapping=None):
        'Checks if the key press matches this shortcut.'

        # First test: key equality.
        if key <> self.key:
            return False

        # Second test: modifier compatibility.
        for kmod, modName in keyboard.KMOD_NAMES:
            if self.mods & kmod:
                if not (modifiers & kmod):
                    return False
            elif modifiers & kmod:
                return False
        # Testing modifiers as done above ignores left/right status.

        return True

class Hotkeys(Element):
    '''
    Traps any keys which occur in the mapping and carries out some action on
    them. Note that this observes only, but does not stop the keypress events
    from passing through.
    '''
    def __init__(self, app, mapping, onActivated=None):
        super(Hotkeys, self).__init__(app)
        self.mapping = mapping
        self.onActivated = Event()
        if onActivated is not None:
            self.onActivated.addListener(onActivated)

    def processEvent(self, event):
        if event.type == pygame.KEYDOWN:
            try:
                action = self.mapping[event.key]
            except KeyError:
                return event

            # Process this hotkey.
            self.onActivated.execute(action)
        return event

    def __repr__(self):
        if isinstance(self.key, int):
            key = self.key
        else:
            # Calculate the physical key.
            for key, value in self.mapping.iteritems():
                if value == self.key:
                    break
            else:
                return '??'

        return keyboard.shortcutName(key, 0)

class MappingHotkey(Element):
    def __init__(self, app, key, mapping=None):
        '''
        key may be a physical pygame key or a string where
            mapping[pygameKey] = key
        '''
        super(MappingHotkey, self).__init__(app)
        self.key = key
        self.onActivated = Event()

        if mapping is None:
            mapping = {}
        self.mapping = mapping

    def processEvent(self, event):
        if event.type == pygame.KEYDOWN:
            if self.mapping.get(event.key, None) == self.key:
                self.onActivated.execute()
                return None
            if event.key == self.key:
                self.onActivated.execute()
                return None
        return event

    def __repr__(self):
        if isinstance(self.key, int):
            key = self.key
        else:
            # Calculate the physical key.
            for key, value in self.mapping.iteritems():
                if value == self.key:
                    break
            else:
                return '??'

        return keyboard.shortcutName(key, 0)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging

log = logging.getLogger(__name__)


class MessageClass(type):
    '''
    This metaclass is provided for convenience. It checks if you have provided
    a single string for the .fields class attribute and if so converts it to a
    tuple.
    '''
    def __new__(cls, name, bases, dictn):
        # Test for a single string cls.fields and make tuple.
        if isinstance(dictn.get('fields'), str):
            dictn['fields'] = (dictn['fields'],)

        # Do the real building.
        return type.__new__(cls, name, bases, dictn)


class MessageBase(object):
    '''
    Allows message classes to be defined by:
        class CustomMessage(Message):
            fields = ('myField', 'otherField')
    And then instantiated by:
        x = CustomMessage(17, otherField=2)
    Which will create an object where:
        x.myField == 17
        x.otherField == 2
    '''

    def __init__(self, *args, **kwargs):
        if len(args) > len(self.fields):
            raise TypeError(
                '%s takes at most %d arguments (%d given)' % (
                    type(self), len(self.fields), len(args)))

        # Simply get the args and store them as attributes of self.
        for i in xrange(len(args)):
            k = self.fields[i]
            v = args[i]
            if k in kwargs:
                raise TypeError(
                    'got multiple values for keyword argument %r' % (k,))
            kwargs[k] = v

        for k, v in kwargs.iteritems():
            setattr(self, k, v)

    def __repr__(self):
        def get(k):
            try:
                return repr(getattr(self, k))
            except AttributeError:
                return '???'
        args = []
        for k in self.fields:
            args.append(get(k))
        return '%s(%s)' % (type(self).__name__, ', '.join(args))

    _trace_level = 0

    def _trace_send(self, target, why):
        log.warning('{}[{}]: {}{}-> {}'.format(
            self.__class__.__name__, hex(id(self))[:8],
            '  ' * self._trace_level, '({})'.format(why) if why else '',
            target))

    def _trace_enter(self, plug):
        self._trace_level += 1
        log.warning('{}[{}]: {}{}'.format(
            self.__class__.__name__, hex(id(self))[:8],
            '  ' * self._trace_level, plug))

    def _trace_exit(self, plug):
        self._trace_level -= 1

    @classmethod
    def handler(cls, fn):
        '''
        Decorator for methods which handle this message type, particularly
        useful when defining MessageConsumer subclasses.
        '''
        makeHandler(fn, cls)
        return fn


class Message(MessageBase):
    __metaclass__ = MessageClass


class UnhandledMessage(Exception):
    pass


def isHandler(fn):
    return bool(getattr(fn, '_handles_msgs', None))


def makeHandler(method, message):
    if not isHandler(method):
        method._handles_msgs = []
    method._handles_msgs.append(message)


class MessageConsumer(object):
    '''
    Base class for any class that may handle different messages based on the
    message type.
    '''

    def __init__(self, *args, **kwargs):
        super(MessageConsumer, self).__init__(*args, **kwargs)

        self.messageDispatchers = {}
        for k in dir(self):
            try:
                v = getattr(self, k)
            except AttributeError:
                # This might be a property which relies on values not yet set
                continue
            if not isHandler(v):
                continue
            for message in v._handles_msgs:
                if message in self.messageDispatchers:
                    raise KeyError(
                        'handler already defined for %s' % (message,))
                self.messageDispatchers[message] = v

    def consumeMsg(self, msg):
        '''
        Sends this message to the appropriate handler function in this class.
        '''
        dispatcher = self.messageDispatchers.get(
            msg.__class__, self.defaultHandler)
        return dispatcher(msg)

    def defaultHandler(self, msg):
        '''
        Called when no other handler matches a message passed to consumeMsg.
        '''
        raise UnhandledMessage(
            '%s does not handle %s' % (self.__class__.__name__, type(msg)))



# Based on https://people.gnome.org/~johan/lsprofcalltree.py
# lsprofcalltree.py: lsprof output which is readable by kcachegrind
# David Allouche
# Jp Calderone & Itamar Shtull-Trauring
# Johan Dahlin
# Modified by Joshua D Bartlett


def get_class_name(code):
    try:
        f = open(code.co_filename, 'rU')
    except IOError:
        return None

    cls = None
    num = 0
    for line in f:
        num += 1
        indent = len(line) - len(line.lstrip())
        if num == code.co_firstlineno:
            break
        if line.lstrip() != '' and indent == 0:
            if line.startswith('class '):
                i = j = len('class ')
                while line[j].isalnum() or line[j] == '_':
                    j += 1
                cls = line[i:j]
            else:
                cls = 0
    else:
        return None

    if indent == 4 and cls:
        return cls

    return None


def label(code):
    if isinstance(code, str):
        return '%s (built-in)' % (code,)
    else:
        classname = get_class_name(code)
        if classname:
            return '%s.%s %s:%d' % (
                classname, code.co_name, code.co_filename, code.co_firstlineno)
        else:
            return '%s %s:%d' % (
                code.co_name, code.co_filename, code.co_firstlineno)


class KCacheGrindOutputter(object):
    def __init__(self, profiler):
        self.data = profiler.getstats()
        self.out_file = None

    def output(self, out_file):
        self.out_file = out_file
        print >> out_file, 'events: Ticks'
        self._print_summary()
        for entry in self.data:
            self._entry(entry)

    def _print_summary(self):
        max_cost = 0
        for entry in self.data:
            totaltime = int(entry.totaltime * 1000)
            max_cost = max(max_cost, totaltime)
        print >> self.out_file, 'summary: %d' % (max_cost,)

    def _entry(self, entry):
        out_file = self.out_file

        code = entry.code
        if isinstance(code, str):
            print >> out_file, 'fi=~'
        else:
            print >> out_file, 'fi=%s' % (code.co_filename,)
        print >> out_file, 'fn=%s' % (label(code),)

        inlinetime = int(entry.inlinetime * 1000)
        if isinstance(code, str):
            print >> out_file, '0 ', inlinetime
        else:
            print >> out_file, '%d %d' % (code.co_firstlineno, inlinetime)

        # recursive calls are counted in entry.calls
        if entry.calls:
            calls = entry.calls
        else:
            calls = []

        if isinstance(code, str):
            lineno = 0
        else:
            lineno = code.co_firstlineno

        for subentry in calls:
            self._subentry(lineno, subentry)
        print >> out_file

    def _subentry(self, lineno, subentry):
        out_file = self.out_file
        code = subentry.code
        print >> out_file, 'cfn=%s' % (label(code),)
        if isinstance(code, str):
            print >> out_file, 'cfi=~'
            print >> out_file, 'calls=%d 0' % (subentry.callcount,)
        else:
            print >> out_file, 'cfi=%s' % (code.co_filename,)
            print >> out_file, 'calls=%d %d' % (
                subentry.callcount, code.co_firstlineno)

        totaltime = int(subentry.totaltime * 1000)
        print >> out_file, '%d %d' % (lineno, totaltime)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.framework import framework, elements
from trosnoth.utils.event import Event
import pygame

class UnobtrusiveValueGetter(framework.Element):
    '''
    Gets some result from the use without obstructing normal use of the
    program.
    '''
    def __init__(self, app, pos, message, font, colour=(100, 100, 0),
            timer=None):
        framework.Element.__init__(self, app)
        self.drawIt = False
        self.text = elements.TextElement(app, message, font, pos, colour)
        if message:
            self.drawIt = True
        self.onGotValue = Event()

        if timer != None:
            self.timeRemaining = timer
        else:
            self.timeRemaining = None

    def draw(self, screen):
        if self.drawIt:
            self.text.draw(screen)

    def timerOver(self):
        self.drawIt = False

    def tick(self, deltaT):
        if self.timeRemaining is not None:
            self.timeRemaining -= deltaT
            if self.timeRemaining <=0:
                self.timeRemaining = None
                self.timerOver()

    def processEvent(self, event):
        if event.type == pygame.KEYDOWN:
            value = self.getValue(event.key)
            if value is not None:
                self.onGotValue.execute(value)
                return None
        return event

    def getValue(self, key):
        raise NotImplementedError, "This should be implemented in subclasses"

class NumberGetter(UnobtrusiveValueGetter):
    '''Gets numbers.'''
    def getValue(self, key):
        numKeys = {pygame.K_0: 0,
                   pygame.K_1: 1,
                   pygame.K_2: 2,
                   pygame.K_3: 3,
                   pygame.K_4: 4,
                   pygame.K_5: 5,
                   pygame.K_6: 6,
                   pygame.K_7: 7,
                   pygame.K_8: 8,
                   pygame.K_9: 9,
                   pygame.K_MINUS: 10,
                   pygame.K_KP0: 0,
                   pygame.K_KP1: 1,
                   pygame.K_KP2: 2,
                   pygame.K_KP3: 3,
                   pygame.K_KP4: 4,
                   pygame.K_KP5: 5,
                   pygame.K_KP6: 6,
                   pygame.K_KP7: 7,
                   pygame.K_KP8: 8,
                   pygame.K_KP9: 9,
                   pygame.K_KP_MINUS: 10,
                   pygame.K_RETURN: 0,
                   pygame.K_KP_ENTER: 0}

        return numKeys.get(key, None)

class YesNoGetter(UnobtrusiveValueGetter):
    '''Gets a boolean (y/n)'''
    def getValue(self, key):
        if key == pygame.K_y:
            return True
        elif key == pygame.K_n:
            return False
        else:
            return None




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import os
import pygame

from trosnoth.gui.framework import framework
from trosnoth.gui.framework.listbox import ListBox
from trosnoth.gui.framework.elements import TextElement
from trosnoth.gui.framework.tab import Tab
from trosnoth.gui.common import ScaledLocation, ScaledArea
from trosnoth.trosnothgui.common import button
from trosnoth.utils.event import Event

from trosnoth.data import getPath, user, makeDirs, themes

class ThemeTab(Tab, framework.TabFriendlyCompoundElement):
    def __init__(self, app, onClose=None, onRestart=None):
        super(ThemeTab, self).__init__(app, 'Themes')

        self.onClose = Event()
        if onClose is not None:
            self.onClose.addListener(onClose)
        self.onRestart = Event()
        if onRestart is not None:
            self.onRestart.addListener(onRestart)

        font = self.app.screenManager.fonts.menuFont
        colours = self.app.theme.colours

        self.inactiveTheme = False
        self.originalTheme = app.theme.name

        # Static text
        self.staticText = [
            TextElement(self.app, 'theme information:', font,
                    ScaledLocation(960, 250, 'topright'),
                    colours.headingColour),
            TextElement(self.app, 'theme contents:', font,
                    ScaledLocation(960, 390, 'topright'), colours.headingColour)
        ]

        # Dynamic text
        self.feedbackText = TextElement(self.app, 'Your current theme: %s' %
                (app.theme.name,), font, ScaledLocation(512, 200, 'midtop'),
                colours.startButton)
        self.listHeaderText = TextElement(self.app, 'available themes:', font,
                ScaledLocation(70, 250), colours.headingColour)
        self.themeNameText = TextElement(self.app, 'Default Theme', font,
                ScaledLocation(960, 290, 'topright'), colours.startButton)
        self.themeAuthorText = TextElement(self.app,
                'created by: Trosnoth Team', font,
                ScaledLocation(960, 330, 'topright'), colours.startButton)

        self.contents = []

        numContents = 4
        for yPos in range(430, 430 + numContents * 40 + 1, 40):
            self.contents.append(TextElement(self.app, '', font,
                    ScaledLocation(960, yPos, 'topright'), colours.startButton))

        self.dynamicText = [self.feedbackText, self.listHeaderText,
                self.themeNameText, self.themeAuthorText] + self.contents

        # Theme list
        self.themeList = ListBox(self.app, ScaledArea(70, 290, 400, 290), [],
                font, colours.listboxButtons)
        self.themeList.onValueChanged.addListener(self.updateSidebar)

        # Text buttons
        self.useThemeButton = button(app, 'use selected theme',
                self.applyTheme, (0, -125), 'midbottom')
        self.refreshButton = button(app, 'refresh', self.populateList,
                (-100, -75), 'midbottom')
        self.cancelButton = button(app, 'cancel', self.backToMain, (100, -75),
                'midbottom')
        self.restartButton = button(app, 'restart Trosnoth', self.restart,
                (0, -125), 'midbottom')

        self.buttons = [self.useThemeButton, self.refreshButton,
                self.cancelButton, self.restartButton]

        # Combine the elements
        self.elements = self.staticText + self.dynamicText + self.buttons + [
                self.themeList]

        self.contentTypes = {"sprites": "sprite", "blocks": "map block",
                "fonts": "font", "startupMenu": "backdrop"}

        # Populate the list of replays
        self.populateList()

    def populateList(self):

        defaultTheme = {"name": "Default Theme", "filename": "default",
                "author": "Trosnoth Team", "content": None,
                "source": "internal"}

        # Clear out the sidebar
        self.themeNameText.setText('')
        self.themeAuthorText.setText('')
        for element in self.contents:
            element.setText('')
        self.useThemeButton.setText('')
        self.restartButton.setText('')
        self.listHeaderText.setText('available themes:')
        self.themeList.index = -1

        userThemeDir = getPath(user, 'themes')
        internalThemeDir = getPath(themes)
        makeDirs(userThemeDir)

        themeList = []

        # Get a list of internal themes
        for dirName in os.listdir(internalThemeDir):
            if os.path.isdir(os.path.join(internalThemeDir, dirName)):
                themeList.append("i/%s" % dirName)

        # Get a list of user-defined themes
        for dirName in os.listdir(userThemeDir):
            if os.path.isdir(os.path.join(userThemeDir, dirName)):
                # Internal themes overrule user-defined themes
                if "i/" + dirName not in themeList and dirName != "default":
                    themeList.append("u/%s" % dirName)

        # Assume all themes are valid for now
        validThemes = themeList[:]

        self.themeInfo = {}

        for themeName in themeList:
            themeInfo = {"content": {}}

            if themeName.startswith("i/"):
                themeInfo['source'] = 'internal'
                directory = internalThemeDir
            else:
                themeInfo['source'] = 'user-defined'
                directory = userThemeDir

            themeNameList = themeName
            themeName = themeName[2:]

            themeInfo['filename'] = themeName[2:]

            anyContent = False

            for contentType in self.contentTypes.keys():
                if themeInfo['source'] == 'internal':
                    contentDir = os.path.join(directory, themeName, contentType)
                else:
                    contentDir = os.path.join(directory, themeName, contentType)

                if not os.path.isdir(contentDir):
                    continue
                else:
                    fileCount = len([f for f in os.listdir(contentDir) if
                            os.path.isfile(os.path.join(contentDir, f))])
                    if fileCount > 0:
                        anyContent = True
                        themeInfo["content"][contentType] = fileCount

            if not anyContent:
                validThemes.remove(themeNameList)
                continue

            infoFile = os.path.join(directory, themeName, "info.txt")
            if os.path.isfile(infoFile):
                infoFile = open(infoFile)
                infoContents = infoFile.readlines()
            else:
                infoContents = []

            if len(infoContents) >= 2:
                themeInfo["author"] = infoContents[1].strip()
            else:
                themeInfo["author"] = None

            if len(infoContents) >= 1:
                themeInfo["name"] = infoContents[0].strip()
            else:
                themeInfo["name"] = themeName

            self.themeInfo[themeName] = themeInfo

        self.themeInfo["default"] = defaultTheme

        # Sort the themes alphabetically
        items = [(v['filename'], n) for n, v in self.themeInfo.iteritems()]
        items.sort()
        items = [n for v,n in items]
        self.themeList.setItems(items)

        if len(self.themeInfo) == 1:
            self.listHeaderText.setText("1 available theme:")
            self.themeList.index = 0
            self.updateSidebar(0)
        else:
            self.listHeaderText.setText("%d available themes:" %
                    len(self.themeInfo))

    def updateSidebar(self, listID):
        # Update the details on the sidebar
        displayName = self.themeList.getItem(listID)

        themeInfo = self.themeInfo[displayName]

        if themeInfo['source'] == "internal":
            self.themeNameText.setText(themeInfo['name'] + " (built-in)")
        else:
            self.themeNameText.setText(themeInfo['name'])
        if themeInfo['author'] is not None:
            self.themeAuthorText.setText("by " + themeInfo['author'])
        else:
            self.themeAuthorText.setText("(creator unknown)")

        if themeInfo['content'] is None:
            for element in self.contents:
                element.setText('')
            self.contents[0].setText('N/A')
        else:
            count = 0
            for k, v in themeInfo['content'].iteritems():
                suffix = ""
                if v != 1:
                    suffix = "s"

                text = "%d %s%s" % (v, self.contentTypes[k], suffix)
                self.contents[count].setText(text)
                count += 1

        self.restartButton.setText("")
        if self.app.displaySettings.theme != displayName:
            if displayName == self.originalTheme:
                self.useThemeButton.setText("revert back to this theme")
            else:
                self.useThemeButton.setText("use this theme")
        else:
            self.useThemeButton.setText("")

            if self.inactiveTheme:
                self.restartButton.setText('restart Trosnoth')

    def applyTheme(self):
        themeName = self.themeList.getItem(self.themeList.index)

        self.app.displaySettings.theme = themeName
        self.app.displaySettings.save()

        self.useThemeButton.setText("")

        if themeName != self.originalTheme:
            self.feedbackText.setText("You must restart Trosnoth for the "
                    "'%s' theme to take effect." % themeName)
            self.restartButton.setText('restart Trosnoth')
            self.inactiveTheme = True
        else:
            self.feedbackText.setText('Your current theme: %s' %
                    (self.app.theme.name,))
            self.restartButton.setText("")
            self.inactiveTheme = False

    def restart(self):
        self.onRestart.execute()

    def backToMain(self):
        if self.inactiveTheme:
            self.feedbackText.setText("Your current theme: %s "
                    "(restart required)" % self.app.displaySettings.theme)

        self.onClose.execute()

    def draw(self, surface):
        super(ThemeTab, self).draw(surface)
        width = max(int(3*self.app.screenManager.scaleFactor), 1)
        scalePoint = self.app.screenManager.placePoint
        pygame.draw.line(surface, self.app.theme.colours.mainMenuColour,
                scalePoint((512,260)), scalePoint((512, 580)), width)




import logging
import weakref

from twisted.internet import reactor, task

log = logging.getLogger('utils.twist')


class WeakCallLater(object):
    '''
    Calls twisted.internet.reactor.callLater, but only maintains a weak
    reference to the object whose method is to be called. If the object is
    garbage collected, the callLater is cancelled.
    '''
    def __init__(self, delay, obj, attr, *args, **kw):
        self.obj = weakref.ref(obj, self._collect)
        self.attr = attr
        self.delayedCall = reactor.callLater(delay, self._call, *args, **kw)
        self.alive = True
        self._str = str(obj)

    def cancel(self):
        self.delayedCall.cancel()

    def getTime(self):
        return self.delayedCall.getTime()

    def delay(self, secondsLater):
        self.delayedCall.delay(secondsLater)

    def reset(self, secondsFromNow):
        self.delayedCall.reset(secondsFromNow)

    def active(self):
        return self.delayedCall.active()

    def _call(self, *args, **kwargs):
        obj = self.obj()
        if obj is None:
            log.warning(
                'callLater called after garbage collection: %s.%s' %
                (self._str, self.attr))
        else:
            return getattr(obj, self.attr)(*args, **kwargs)

    def _collect(self, r):
        log.info('WeakCallLater: Collecting %s' % (self._str,))
        if self.delayedCall.active():
            self.delayedCall.cancel()
        self.alive = False


class WeakLoopingCall(object):
    '''
    Wraps a twisted.internet.task.LoopingCall, but only maintains a weak
    reference to the object whose method is to be called. If the object is
    garbage collected, the looping call is stopped.
    '''
    def __init__(self, obj, attr, *a, **kw):
        self.alive = True
        self.obj = weakref.ref(obj, self._collect)
        self.attr = attr
        self.loop = task.LoopingCall(self._call, *a, **kw)
        self._str = str(obj)

    @property
    def running(self):
        return self.loop.running

    def start(self, interval, now=True):
        self.loop.start(interval, now)

    def stop(self):
        if self.alive:
            self.loop.stop()

    def _collect(self, r):
        log.info('WeakLoopingCall: Collecting %s' % (self._str,))
        self.alive = False
        if self.loop.running:
            self.loop.stop()

    def _call(self, *args, **kwargs):
        if self.alive:
            obj = self.obj()
            if obj is None:
                log.warning(
                    'Looping call called after garbage collection: %s.%s' %
                    (self._str, self.attr))
            else:
                getattr(obj, self.attr)(*args, **kwargs)



import logging
import struct

from twisted.internet import defer
from twisted.internet.protocol import Factory
from twisted.protocols.basic import Int32StringReceiver

from trosnoth.model.hub import Hub, Node
from trosnoth.utils import netmsg

log = logging.getLogger(__name__)


AGENT_REQUEST_ID_LIMIT = (1 << 16)

CLIENT_CONNECT_AGENT = 'c'
CLIENT_DISCONNECT_AGENT = 'd'
CLIENT_REQUEST = 'm'

SERVER_GENERAL_MESSAGE = 'g'
SERVER_TARGETED_MESSAGE = 't'
SERVER_CONNECTED_AGENT = 'c'
SERVER_DISCONNECTED_AGENT = 'd'


class MsgClientProtocol(Int32StringReceiver, Hub):
    messages = netmsg.MessageCollection()
    greeting = None     # Override this in subclasses

    def connectionMade(self):
        self.agentRequests = {}
        self.nextAgentRequestId = 0
        self.sendString(self.greeting)
        self.agentIds = set()

    def connectionLost(self, reason=None):
        if self.node:
            while self.agentIds:
                agentId = self.agentIds.pop()
                self.node.agentDisconnected(agentId)
        else:
            self.agentIds.clear()

    def stringReceived(self, string):
        kind, body = string[:1], string[1:]
        try:
            if kind == SERVER_GENERAL_MESSAGE:
                self.receivedGeneralMessageString(body)
            elif kind == SERVER_TARGETED_MESSAGE:
                self.receivedTargetedMessageString(body)
            elif kind == SERVER_CONNECTED_AGENT:
                self.receivedAgentConnectedString(body)
            elif kind == SERVER_DISCONNECTED_AGENT:
                self.receivedAgentDisconnectedString(body)
            else:
                self.receivedBadString(string)
        except Exception:
            log.exception('Error responding to received message')

    def receivedBadString(self, string):
        log.warning(
            'Got unexpected string from remote end, dropping connection.')
        self.transport.abortConnection()

    def receivedGeneralMessageString(self, string):
        msg = self.buildMessage(string)
        if msg:
            self.gotGeneralMsg(msg)
        else:
            self.receivedBadString(string)

    def gotGeneralMsg(self, msg):
        if self.node:
            self.node.gotServerCommand(msg)

    def receivedTargetedMessageString(self, string):
        agentId, body = string[:2], string[2:]

        msg = self.buildMessage(body)
        if not msg:
            self.receivedBadString(string)
        elif self.node:
            self.node.gotMessageToAgent(agentId, msg)

    def buildMessage(self, string):
        try:
            return self.messages.buildMessage(string)
        except netmsg.NetworkMessageError:
            return None

    def receivedAgentConnectedString(self, string):
        reqId, agentId = string[:2], string[2:]
        try:
            d = self.agentRequests.pop(reqId)
        except KeyError:
            return
        self.agentIds.add(agentId)
        d.callback(agentId)

    def receivedAgentDisconnectedString(self, agentId):
        self.agentIds.discard(agentId)
        if self.node:
            self.node.agentDisconnected(agentId)

    def connectNewAgent(self):
        if len(self.agentRequests) >= AGENT_REQUEST_ID_LIMIT:
            raise OverflowError('Too many concurrent requests for new agents')
        while True:
            reqId = struct.pack('!H', self.nextAgentRequestId)
            self.nextAgentRequestId += 1
            self.nextAgentRequestId %= AGENT_REQUEST_ID_LIMIT
            if reqId not in self.agentRequests:
                break
        d = defer.Deferred()
        self.agentRequests[reqId] = d
        self.sendString(CLIENT_CONNECT_AGENT + reqId)

        return d

    def disconnectAgent(self, agentId):
        self.sendString(CLIENT_DISCONNECT_AGENT + agentId)

    def sendRequestToGame(self, agentId, msg):
        self.sendString(CLIENT_REQUEST + agentId + msg.pack())


class MsgServerProtocol(Int32StringReceiver, Node):
    def connectionMade(self):
        self.initialised = False
        self.greeting = None

    def connectionLost(self, reason=None):
        if self.initialised:
            self.factory.connectionLost(self, reason)

    def stringReceived(self, string):
        if self.initialised:
            self.receivedMessageString(string)
        else:
            if self.factory.checkGreeting(string):
                self.initialised = True
                self.greeting = string
                self.factory.connectionEstablished(self)
            else:
                self.transport.abortConnection()

    def receivedMessageString(self, string):
        kind, body = string[:1], string[1:]
        if kind == CLIENT_REQUEST:
            self.receivedRequestString(body)
        elif kind == CLIENT_CONNECT_AGENT:
            self.receivedConnectAgentString(body)
        elif kind == CLIENT_DISCONNECT_AGENT:
            self.receivedDisconnectAgentString(body)
        else:
            self.factory.gotBadString(self, string)

    def receivedRequestString(self, string):
        agentId, body = string[:2], string[2:]
        try:
            msg = self.factory.messages.buildMessage(body)
        except netmsg.NetworkMessageError:
            self.factory.gotBadString(self, string)
            return
        self.hub.sendRequestToGame(self.decodeAgentId(agentId), msg)

    @defer.inlineCallbacks
    def receivedConnectAgentString(self, reqId):
        agentId = yield self.hub.connectNewAgent()
        self.sendString(
            SERVER_CONNECTED_AGENT + reqId + self.encodeAgentId(agentId))

    def receivedDisconnectAgentString(self, agentId):
        self.hub.disconnectAgent(self.decodeAgentId(agentId))

    def encodeAgentId(self, agentId):
        '''
        Takes an agentId in the form given by our hub and converts it to a form
        the network can understand.
        '''
        return struct.pack('!H', agentId)

    def decodeAgentId(self, agentId):
        '''
        Takes an agentId in the form given by the network and converts it into
        a form our hub will understand.
        '''
        return struct.unpack('!H', agentId)[0]

    def gotServerCommand(self, msg):
        self.sendString(SERVER_GENERAL_MESSAGE + msg.pack())

    def gotMessageToAgent(self, agentId, msg):
        self.sendString(
            SERVER_TARGETED_MESSAGE + struct.pack('!H', agentId) + msg.pack())

    def agentDisconnected(self, agentId):
        self.sendString(SERVER_DISCONNECTED_AGENT + agentId)


class MsgServer(Factory):
    '''
    Base class for servers which receive packed NetworkMessage objects from the
    network.
    '''
    protocol = MsgServerProtocol
    messages = netmsg.MessageCollection()

    def checkGreeting(self, greeting):
        '''
        Must check whether the given greeting string is allowed for this
        server, and return true or false accordingly.
        '''
        return False

    def connectionEstablished(self, protocol):
        '''
        Called after a client has connected and sent an appropriate greeting
        string.
        '''
        pass

    def connectionLost(self, protocol, reason):
        pass

    def gotBadString(self, protocol, string):
        '''
        Called when a client has sent a string that cannot be decoded by our
        known message types.
        '''
        log.warning('Client sent bad message string')
        protocol.transport.abortConnection()



import logging
from math import atan2

import pygame

from trosnoth.gui.framework import framework
from trosnoth.messages import UpdatePlayerStateMsg, AimPlayerAtMsg, ShootMsg

log = logging.getLogger(__name__)


class TickInstructions(object):
    def __init__(self):
        self.state = {}
        self.angleFacing = None
        self.ghostThrust = None
        self.firedGun = False

    def updateState(self, key, value):
        self.state[key] = value

    def lookAt(self, angle, thrust):
        self.angleFacing = angle
        self.ghostThrust = thrust

    def applyTo(self, player):
        for key, value in self.state.iteritems():
            if player.getState(key) == value:
                if not value:
                    continue
                player.updateState(key, False)
            player.updateState(key, value)

        if self.angleFacing is not None:
            player.lookAt(self.angleFacing, self.ghostThrust)

        if self.firedGun:
            player.weaponDischarged()


class PlayerInterface(framework.Element):
    '''Interface for controlling a player.'''

    # The virtual keys we care about.
    state_vkeys = frozenset(['left', 'right', 'jump', 'down'])

    NO_GHOST_THRUST_MOUSE_RADIUS = 30       # pixels
    FULL_GHOST_THRUST_MOUSE_RADIUS = 150    # pixels

    def __init__(self, app, gameInterface):
        super(PlayerInterface, self).__init__(app)

        world = gameInterface.world
        self.gameInterface = gameInterface
        self.keyMapping = gameInterface.keyMapping

        self.receiving = True

        self.world = world
        self.worldGui = gameInterface.gameViewer.worldgui

        self.mousePos = (0, 0)
        pygame.mouse.get_rel()

        # Make sure the viewer is focusing on this player.
        self.gameInterface.gameViewer.setTarget(self.playerSprite)

    def getTickId(self):
        return self.world.lastTickId

    def stop(self):
        pass

    @property
    def player(self):
        return self.gameInterface.localState.player

    @property
    def playerSprite(self):
        return self.worldGui.getPlayerSprite(self.player.id)

    def tick(self, deltaT):
        pos = pygame.mouse.get_pos()
        spritePos = self.playerSprite.rect.center
        self.mousePos = (pos[0] - spritePos[0], pos[1] - spritePos[1])
        self.updatePlayerViewAngle()
        if self.player.canShoot():
            if self.player.mgFiring:
                if self.player.machineGunner:
                    self._fireShot()

    def updatePlayerViewAngle(self):
        '''Updates the viewing angle of the player based on the mouse pointer
        being at the position pos. This gets its own method because it needs
        to happen as the result of a mouse motion and of the viewManager
        scrolling the screen.'''
        dx, dy = self.mousePos
        pos = (self.playerSprite.rect.center[0] + dx,
               self.playerSprite.rect.center[1] + dy)

        if self.playerSprite.rect.collidepoint(pos):
            return

        # Angle is measured clockwise from vertical.
        theta = atan2(dx, -dy)
        dist = (dx ** 2 + dy ** 2) ** 0.5

        # Calculate a thrust value based on distance.
        if dist < self.NO_GHOST_THRUST_MOUSE_RADIUS:
            thrust = 0.0
        elif dist > self.FULL_GHOST_THRUST_MOUSE_RADIUS:
            thrust = 1.0
        else:
            span = (
                self.FULL_GHOST_THRUST_MOUSE_RADIUS -
                self.NO_GHOST_THRUST_MOUSE_RADIUS)
            thrust = (dist - self.NO_GHOST_THRUST_MOUSE_RADIUS) / span

        self.gameInterface.sendRequest(
            AimPlayerAtMsg(theta, thrust, tickId=self.getTickId()))

    def processEvent(self, event):
        '''Event processing works in the following way:
        1. If there is a prompt on screen, the prompt will either use the
        event, or pass it on.
        2. If passed on, the event will be sent back to the main class, for it
        to process whether player movement uses this event. If it doesn't use
        the event, it will pass it back.
        3. If so, the hotkey manager will see if the event means anything to
        it.  If not, that's the end, the event is ignored.
        '''

        # Handle events specific to in-game.
        di = self.gameInterface.detailsInterface
        if self.player:
            if event.type == pygame.KEYDOWN:
                self.processKeyEvent(event, True)
            elif event.type == pygame.KEYUP:
                self.processKeyEvent(event, False)
            elif event.type == pygame.MOUSEBUTTONDOWN:
                if event.button == 1:
                    if self.player.dead:
                        di.doAction('respawn')
                    elif self.player.machineGunner:
                        self.player.mgFiring = True
                    elif di.trajectoryOverlay.isActive():
                        di.doAction('activate upgrade')
                    else:
                        # Fire a shot.
                        self._fireShot()
                elif event.button == 3:
                    di.trajectoryOverlay.setEnabled(True)
            elif event.type == pygame.MOUSEBUTTONUP:
                if event.button == 1:
                    if self.player.machineGunner:
                        self.player.mgFiring = False
                elif event.button == 3:
                    di.trajectoryOverlay.setEnabled(False)

            else:
                return event

    def processKeyEvent(self, event, keyDown):
            try:
                stateKey = self.keyMapping[event.key]
            except KeyError:
                return event

            if stateKey not in self.state_vkeys:
                return event

            self.gameInterface.sendRequest(UpdatePlayerStateMsg(
                keyDown, stateKey=stateKey,
                tickId=self.getTickId()))

    def _fireShot(self):
        '''Fires a shot in the direction the player's currently looking.'''
        self.gameInterface.sendRequest(ShootMsg(self.getTickId()))



import logging
from math import atan2, pi
import random

from twisted.internet import reactor

from trosnoth.ai import AI
from trosnoth.ais.pathfinding import NORTH, SOUTH, EAST, WEST, ORB
from trosnoth.messages import TickMsg
from trosnoth.utils import math, globaldebug
from trosnoth.utils.event import Event
from trosnoth.utils.twist import WeakLoopingCall

log = logging.getLogger(__name__)


SHOOTING = True


class RangerAI(AI):
    nick = 'RangerBot'
    playable = True

    def start(self):
        self.onTick = Event()
        self.recentGoals = {}
        self.clearRecent = None
        self.mainGoal = WinCurrentGame(self, None)

        self.mainGoal.start()

    def disable(self):
        self.mainGoal.stop()
        if self.clearRecent:
            self.clearRecent.cancel()
            self.clearRecent = None

    def showGoalStack(self):
        '''
        For debugging.
        '''
        log.error('Goal stack for %s:', self.player)
        curGoal = self.mainGoal
        while curGoal:
            log.error('  %s', curGoal)
            curGoal = curGoal.subGoal
        log.error('')

    def subGoalStopped(self, goal):
        # Record recent inactive goals so they can be reused if they come up
        # again.
        self.recentGoals[goal] = goal
        if self.clearRecent:
            self.clearRecent.cancel()
            self.clearRecent = None
        self.clearRecent = reactor.callLater(4, self.clearRecentGoals)

    def startingSubGoal(self, goal):
        if goal in self.recentGoals:
            del self.recentGoals[goal]

    def clearRecentGoals(self):
        self.clearRecent = None
        self.recentGoals.clear()

    def checkGoal(self, goal):
        result = self.recentGoals.get(goal, goal)
        result.parent = goal.parent
        return result

    @TickMsg.handler
    def handle_TickMsg(self, msg):
        self.onTick()

        if SHOOTING and self.player.canShoot():
            physics = self.world.physics
            gunRange = physics.shotLifetime * physics.shotSpeed
            shot = None
            for p in self.world.players:
                if p.dead or self.player.isFriendsWith(p):
                    continue
                if p.invisible or p.turret:
                    continue
                if math.distance(self.player.pos, p.pos) < gunRange:
                    # Check if we can shoot without hitting obstacles
                    if shot is None:
                        shot = self.player.createShot()
                    deltaX = p.pos[0] - self.player.pos[0]
                    deltaY = p.pos[1] - self.player.pos[1]
                    obstacle, dX, dY = physics.trimPathToObstacle(
                        shot, deltaX, deltaY, ())
                    if obstacle is None:
                        self.doAimAtPoint(p.pos)
                        self.doShoot()
                        return


class Goal(object):
    '''
    Represents something that the AI is trying to achieve.
    '''

    def __init__(self, ai, parent):
        self.ai = ai
        self.parent = parent
        self.subGoal = None

    def __str__(self):
        return self.__class__.__name__

    def start(self):
        '''
        Called when this goal should begin its work.
        '''
        self.reevaluate()

    def stop(self):
        '''
        Should disable any active components of this goal.
        '''
        if self.subGoal:
            self.subGoal.stop()
            self.subGoal = None

    def setSubGoal(self, goal):
        '''
        If the given goal is already the current sub-goal, does nothing.
        Otherwise, stops the current sub-goal, and starts the given one.
        '''
        if self.subGoal == goal:
            return
        if self.subGoal:
            self.subGoal.stop()
            self.subGoal.parent = None
            self.ai.subGoalStopped(goal)

        if goal:
            goal = self.ai.checkGoal(goal)
            self.subGoal = goal
            self.ai.startingSubGoal(goal)
            goal.start()
        else:
            self.subGoal = None

    def returnToParent(self):
        '''
        Call this method to tell the parent goal that this goal is either
        completed, or no longer relevant.
        '''
        if self.parent:
            reactor.callLater(0, self.parent.returnedFromChild, self)

    def returnedFromChild(self, child):
        '''
        Called by child's returnToParent() method. The default implementation
        checks that the returning child is this goal's subgoal, then calls
        reevaluate().
        '''
        if child is self.subGoal:
            self.reevaluate()
        else:
            self.ai.subGoalStopped(child)

    def reevaluate(self):
        '''
        Called by the default implementations of start() and
        returnedFromChild() to determine what this goal should do next.
        '''
        pass


class WinCurrentGame(Goal):
    def reevaluate(self):
        state = self.ai.world.state

        if not state.inMatch:
            # Lobby or similar
            self.setSubGoal(RunAroundKillingHumans(self.ai, self))
        elif state.isTrosball():
            self.setSubGoal(ScoreTrosballPoint(self.ai, self))
        elif state.becomeNeutralWhenDie():
            # Rabbit hunt
            self.setSubGoal(HuntTheRabbits(self.ai, self))
        else:
            self.setSubGoal(WinStandardTrosnothGame(self.ai, self))


class ScoreTrosballPoint(Goal):
    def reevaluate(self):
        if not self.ai.world.state.isTrosball():
            self.returnToParent()
            return

        # TODO: trosball logic
        self.setSubGoal(RunAroundKillingThings(self.ai, self))


class RunAroundKillingThings(Goal):
    def start(self):
        self.worldState = self.ai.world.state
        self.ai.world.onGameOver.addListener(self.reevaluate)
        self.ai.world.onReset.addListener(self.reevaluate)
        self.checkLoop = WeakLoopingCall(self, 'reevaluate')
        self.checkLoop.start(3)

    def stop(self):
        super(RunAroundKillingThings, self).stop()
        self.checkLoop.stop()
        self.ai.world.onGameOver.removeListener(self.reevaluate)
        self.ai.world.onReset.removeListener(self.reevaluate)

    def reevaluate(self):
        if self.ai.world.state != self.worldState:
            self.returnToParent()
            return

        player = self.ai.player

        if player.dead:
            zone = self.selectZone()
            if zone is None:
                zone = player.getZone()
                if zone is None:
                    zone = random.choice(list(self.ai.world.zones))
            self.setSubGoal(RespawnNearZone(self.ai, self, zone))
            return

        if player.getZone() and self.zoneIsOk(player.getZone()):
            # There are enemies here
            self.setSubGoal(MessAroundInZone(self.ai, self))
            return

        zone = self.selectZone()
        if zone:
            self.setSubGoal(MoveToOrbWhileAlive(self.ai, self, zone))
        else:
            self.setSubGoal(MessAroundInZone(self.ai, self))

    def zoneIsOk(self, zone):
        return any(
            (not p.dead and not self.ai.player.isFriendsWith(p))
            for p in zone.players)

    def selectZone(self):
        options = []
        for zone in self.ai.world.zones:
            if self.zoneIsOk(zone):
                options.append(zone)
        if not options:
            return None
        return random.choice(options)


class RunAroundKillingHumans(RunAroundKillingThings):
    def zoneIsOk(self, zone):
        return any(
            (not p.dead and not self.ai.player.isFriendsWith(p) and not p.bot)
            for p in zone.players)


class HuntTheRabbits(RunAroundKillingThings):
    def zoneIsOk(self, zone):
        return any(
            (
                not p.dead and not self.ai.player.isFriendsWith(p)
                and p.team is not None)
            for p in zone.players)


class WinStandardTrosnothGame(Goal):
    '''
    Win the current game of Trosnoth by capturing all the zones.
    '''

    def start(self):
        self.ai.world.onGameOver.addListener(self.reevaluate)
        self.ai.world.onReset.addListener(self.reevaluate)
        self.checkLoop = WeakLoopingCall(self, 'reevaluate')
        self.checkLoop.start(3)

    def stop(self):
        super(WinStandardTrosnothGame, self).stop()
        self.checkLoop.stop()
        self.ai.world.onGameOver.removeListener(self.reevaluate)
        self.ai.world.onReset.removeListener(self.reevaluate)

    def reevaluate(self, *args, **kwargs):
        '''
        Decide whether to stay in the current zone, or move to another.
        '''

        state = self.ai.world.state
        if (not state.inMatch) or state.becomeNeutralWhenDie():
            self.returnToParent()
            return

        player = self.ai.player
        myZone = player.getZone()

        # 1. If we're defending a disputed zone, stay in the zone
        if myZone and myZone.owner == player.team and myZone.isDisputed():
            self.setSubGoal(DefendZone(self.ai, self, myZone))
            return

        # 2. If we're attacking a capturable zone, stay in the zone
        if (
                myZone and myZone.owner != player.team
                and myZone.isCapturableBy(player.team)):
            self.setSubGoal(CaptureZone(self.ai, self, myZone))
            return

        # 3. Score other zones based on how helpful it would be to be there and
        #    how likely we are to get there in time.

        if player.dead:
            zone = self.getMostUrgentZone()
        else:
            zone = self.getMostLikelyUrgentZone(myZone)

        if zone is None:
            self.returnToParent()
        elif zone.owner == player.team:
            self.setSubGoal(DefendZone(self.ai, self, zone))
        else:
            self.setSubGoal(CaptureZone(self.ai, self, zone))

    def getMostUrgentZone(self):
        bestScore = 0
        bestOptions = []

        for zone in self.ai.world.zones:
            utility = self.getZoneUtility(zone)

            if not [
                    z for z in zone.getAdjacentZones()
                    if z.owner == zone.owner]:
                # This is the last remaining zone
                awesomeness = 5
            else:
                awesomeness = zone.consequenceOfCapture()

            score = utility * awesomeness
            if score == bestScore:
                bestOptions.append(zone)
            elif score > bestScore:
                bestOptions = [zone]
                bestScore = score

        if not bestOptions:
            return None

        return random.choice(bestOptions)

    def getMostLikelyUrgentZone(self, myZone):
        bestScore = 0
        bestOptions = []
        seen = {myZone}
        pending = [(zone, 0.7) for zone in myZone.getUnblockedNeighbours()]
        while pending:
            zone, likelihood = pending.pop(0)
            seen.add(zone)

            utility = self.getZoneUtility(zone)

            if not [
                    z for z in zone.getAdjacentZones()
                    if z.owner == zone.owner]:
                # This is the last remaining zone
                awesomeness = 5
            else:
                awesomeness = zone.consequenceOfCapture()

            score = likelihood * utility * awesomeness
            if score == bestScore:
                bestOptions.append(zone)
            elif score > bestScore:
                bestOptions = [zone]
                bestScore = score

            likelihood *= 0.7
            for other in zone.getUnblockedNeighbours():
                if other not in seen:
                    pending.append((other, likelihood))

        if not bestOptions:
            return None

        return random.choice(bestOptions)

    def getZoneUtility(self, zone):
        player = self.ai.player

        # Count the number of friendly players and players on the most
        # likely enemy team to tag the zone.
        enemy = friendly = 0
        for count, teams in zone.getPlayerCounts():
            if player.team in teams and not friendly:
                friendly = count
            if [t for t in teams if t != player.team] and not enemy:
                enemy = count
            if friendly and enemy:
                break

        if zone.owner == player.team:
            defence = min(3, friendly)
            if enemy == 0:
                utility = 0
            elif enemy > defence:
                # There's a slim chance you could shoot them before they
                # capture the zone.
                utility = 0.2 ** (enemy - defence)
            elif enemy == defence:
                # Being here will stop it being tagged
                utility = 1
            else:
                # There's a slim chance the enemy might shoot them
                utility = 0.2 ** (friendly - enemy)
        elif not zone.adjacentToAnotherZoneOwnedBy(player.team):
            # Cannot capture, have no adjacent zones
            utility = 0
        else:
            defence = min(3, enemy)
            if friendly > defence:
                # Already capturable, but there's a slim chance teammates
                # might die.
                utility = 0.2 ** (friendly - defence)
            elif friendly == defence:
                # Being here will enable the zone tag
                utility = 1
            else:
                # There's a slim chance you could shoot them and capture
                utility = 0.2 ** (friendly - enemy)
        return utility


class ZoneMixin(Goal):
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return other.zone == self.zone

    def __str__(self):
        return '%s(%s)' % (self.__class__.__name__, self.zone)


class CaptureZone(ZoneMixin, Goal):
    '''
    Respawns if necessary, moves to the given zone, messes around until it's
    capturable, and captures it. If the player dies, respawns and tries again.
    Returns if the zone is captured or becomes uncapturable by virtue of having
    no adjacent zones owned by the team.
    '''

    def __init__(self, ai, parent, zone):
        super(CaptureZone, self).__init__(ai, parent)
        self.zone = zone

    def reevaluate(self):
        player = self.ai.player
        if self.zone.owner == player.team:
            self.returnToParent()
            return

        if not self.zone.adjacentToAnotherZoneOwnedBy(player.team):
            self.returnToParent()
            return

        if player.dead:
            self.setSubGoal(RespawnNearZone(self.ai, self, self.zone))
            return

        playerZone = player.getZone()
        if (
                playerZone == self.zone
                and not self.zone.isCapturableBy(player.team)):
            self.setSubGoal(MessAroundInZone(self.ai, self))
        else:
            self.setSubGoal(MoveToOrbWhileAlive(self.ai, self, self.zone))


class DefendZone(ZoneMixin, Goal):
    '''
    Respawns if necessary, moves to the given zone and messes around there.
    If the player dies, respawns and continues. Returns if the zone is
    captured or neutralised.
    '''

    def __init__(self, ai, parent, zone):
        super(DefendZone, self).__init__(ai, parent)
        self.zone = zone

    def reevaluate(self):
        player = self.ai.player
        if self.zone.owner != player.team:
            self.returnToParent()
            return

        if player.dead:
            self.setSubGoal(RespawnNearZone(self.ai, self, self.zone))
            return

        playerZone = player.getZone()
        if playerZone == self.zone:
            self.setSubGoal(MessAroundInZone(self.ai, self))
        else:
            self.setSubGoal(MoveToOrbWhileAlive(self.ai, self, self.zone))


class MessAroundInZone(Goal):
    '''
    Wander around in the zone the player is in at the time of construction.
    Mostly random, but has a slight tendency to move towards the orb.
    Aborts if the player dies or leaves the zone.
    '''

    def __init__(self, ai, parent):
        super(MessAroundInZone, self).__init__(ai, parent)
        self.zone = self.ai.player.getZone()
        self.action = None

    def start(self):
        self.ai.onTick.addListener(self.tick)
        super(MessAroundInZone, self).start()

    def stop(self):
        super(MessAroundInZone, self).stop()
        self.ai.onTick.removeListener(self.tick)

    def reevaluate(self):
        player = self.ai.player
        if player.dead:
            self.action = None
            self.returnToParent()
            return

        zone = player.getZone()
        if zone != self.zone:
            self.action = None
            self.returnToParent()
            return

        pathFinder = self.ai.world.map.layout.pathFinder
        if pathFinder is None:
            # Can't do anything without a path-finding database loaded
            self.action = None
            log.warning('No pathfinding database loaded')
            self.stop()
            return

        if random.random() < 0.1:
            # Move towards the orb
            block = player.getMapBlock()
            if block.defn.kind in ('top', 'btm'):
                self.action = pathFinder.getOrbAction(self.ai.player)
            elif block.defn.rect.centerx < zone.defn.pos[0]:
                self.action = pathFinder.getExitAction(self.ai.player, EAST)
            else:
                self.action = pathFinder.getExitAction(self.ai.player, WEST)
        else:
            # Move randomly
            self.action = random.choice(
                pathFinder.getAllActions(self.ai.player))

        if self.action is None:
            self.returnToParent()

    def tick(self):
        player = self.ai.player
        if player.dead:
            self.reevaluate()
            return

        if not self.action:
            self.reevaluate()
            if self.action is None:
                return

        done, stateChanges, faceRight = self.action.prepNextStep(player)

        if done:
            self.action = None
            return

        for key, value in stateChanges:
            self.ai._sendStateUpdate(key, value)

        if faceRight and not player.isFacingRight():
            self.ai.doAimAt(pi / 2)
        elif (not faceRight) and player.isFacingRight():
            self.ai.doAimAt(-pi / 2)


class RespawnNearZone(ZoneMixin, Goal):
    '''
    Respawns in a zone that's as close as possible to the given zone.
    '''

    def __init__(self, ai, parent, zone):
        super(RespawnNearZone, self).__init__(ai, parent)
        self.zone = zone

    def start(self):
        self.ai.world.onZoneTagged.addListener(self.zoneTagged)
        super(RespawnNearZone, self).start()

    def stop(self):
        super(RespawnNearZone, self).stop()
        self.ai.world.onZoneTagged.removeListener(self.zoneTagged)

    def zoneTagged(self, *args, **kwargs):
        '''
        We may be waiting to respawn in a zone that's just changed ownership.
        '''
        self.reevaluate()

    def reevaluate(self):
        player = self.ai.player
        if not player.dead:
            self.returnToParent()
            return

        bestZone = self.zone
        zones = [self.zone]
        seen = set()

        while zones:
            zone = zones.pop(0)
            seen.add(zone)
            if player.isZoneRespawnable(zone):
                bestZone = zone
                break
            adjacent = [
                z for z in zone.getUnblockedNeighbours() if z not in seen]
            random.shuffle(adjacent)
            zones.extend(adjacent)

        if player.getZone() == bestZone:
            self.setSubGoal(RespawnWhenPossible(self.ai, self))
        else:
            self.setSubGoal(MoveGhostToZone(self.ai, self, bestZone))


class MoveGhostToZone(ZoneMixin, Goal):
    def __init__(self, ai, parent, zone):
        super(MoveGhostToZone, self).__init__(ai, parent)
        self.zone = zone

    def start(self):
        self.ai.onTick.addListener(self.tick)

        zonePos = self.zone.defn.pos
        playerPos = self.ai.player.pos
        if zonePos == playerPos:
            return
        dx = zonePos[0] - playerPos[0]
        dy = zonePos[1] - playerPos[1]
        theta = atan2(dx, -dy)
        self.ai.doAimAt(theta, 1.0)

    def stop(self):
        super(MoveGhostToZone, self).stop()
        self.ai.onTick.removeListener(self.tick)

    def tick(self):
        if not self.ai.player.dead:
            self.returnToParent()
        elif self.ai.player.getZone() == self.zone:
            self.ai.doAimAt(0.0, 0.0)
            self.returnToParent()


class RespawnWhenPossible(Goal):
    def start(self):
        self.ai.onTick.addListener(self.tick)

    def stop(self):
        super(RespawnWhenPossible, self).stop()
        self.ai.onTick.removeListener(self.tick)

    def tick(self):
        player = self.ai.player
        if not (player.dead and player.inRespawnableZone()):
            # Impossible to respawn
            self.returnToParent()
            return

        if not player.world.abilities.respawn:
            return
        if player.respawnGauge > 0:
            return
        if player.getZone().frozen:
            return

        # Ready to respawn
        self.ai.doRespawn()


class MoveToOrbWhileAlive(ZoneMixin, Goal):

    def __init__(self, ai, parent, zone):
        super(MoveToOrbWhileAlive, self).__init__(ai, parent)
        self.zone = zone
        self.path = None

    def start(self):
        self.ai.onTick.addListener(self.tick)
        if __debug__ and globaldebug.enabled and globaldebug.showPathFinding:
            player = self.ai.world.getPlayer(self.ai.player.id)
            player.onOverlayDebugHook.addListener(self.overlay)

    def stop(self):
        super(MoveToOrbWhileAlive, self).stop()
        self.ai.onTick.removeListener(self.tick)
        if __debug__ and globaldebug.enabled and globaldebug.showPathFinding:
            player = self.ai.world.getPlayer(self.ai.player.id)
            player.onOverlayDebugHook.removeListener(self.overlay)

    def overlay(self, viewManager, screen, sprite):
        '''
        Shows debugging overlay. Enabled using flags in
        trosnoth.utils.globaldebug.
        '''
        if self.path is None or self.path.steps is None:
            return

        import pygame
        from trosnoth.trosnothgui.ingame.utils import mapPosToScreen
        focus = viewManager._focus
        area = viewManager.sRect
        prev = mapPosToScreen(sprite.pos, focus, area)
        for blockDef, scoreKey in self.path.steps:
            if blockDef is None:
                break
            if scoreKey == ORB:
                scoreKey = NORTH if blockDef.kind == 'btm' else SOUTH

            if scoreKey == NORTH:
                pt = mapPosToScreen(blockDef.rect.midtop, focus, area)
            elif scoreKey == SOUTH:
                pt = mapPosToScreen(blockDef.rect.midbottom, focus, area)
            elif scoreKey == EAST:
                pt = mapPosToScreen(blockDef.rect.midright, focus, area)
            elif scoreKey == WEST:
                pt = mapPosToScreen(blockDef.rect.midleft, focus, area)
            else:
                # Shouldn't happen
                pt = mapPosToScreen(blockDef.rect.centre)
            pygame.draw.line(screen, (255, 192, 0), prev, pt, 2)
            prev = pt

    def tick(self):
        player = self.ai.player

        if player.dead:
            self.returnToParent()
            return

        if self.path is None:
            pathFinder = self.ai.world.map.layout.pathFinder
            if pathFinder is None:
                # No path-finding database loaded
                log.warning('No pathfinding database loaded')
                self.stop()
                return

            self.path = pathFinder.getRouteToOrb(self.zone)
            # TODO: if the best path is more than a certain value, use bomber

        done, stateChanges, faceRight = self.path.prepNextStep(player)

        if done:
            self.path = None
            self.returnToParent()
            return

        for key, value in stateChanges:
            self.ai._sendStateUpdate(key, value)

        if faceRight and not player.isFacingRight():
            self.ai.doAimAt(pi / 2)
        elif (not faceRight) and player.isFacingRight():
            self.ai.doAimAt(-pi / 2)


AIClass = RangerAI



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('trosnoth', '0002_auto_20160206_1103'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='achievementprogress',
            unique_together=set([('user', 'achievementId')]),
        ),
        migrations.AlterUniqueTogether(
            name='gameplayer',
            unique_together=set([('game', 'user')]),
        ),
        migrations.AlterUniqueTogether(
            name='playerkills',
            unique_together=set([('game', 'killer', 'killee')]),
        ),
    ]



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pygame

from trosnoth.utils.event import Event
from trosnoth.gui import browser
from trosnoth.gui.framework.framework import CompoundElement
from trosnoth.gui.framework.elements import (SizedPicture, TextElement,
        TextButton)

class NotificationBar(CompoundElement):
    '''
    Displays a bar notifying the user of some information.
    '''

    def __init__(self, app, message, font, area, buttonPos, textPos, url=None,
            textColour=(0,0,0), closeColour=(0,0,0), hoverColour=(255,255,255),
            backColour=(255,150,150), textAnchor='center'):
        super(NotificationBar, self).__init__(app)
        self.area = area
        self.url = url
        self.backColour = backColour

        self.rect = pygame.Rect(0,0,0,0)
        self.background = pygame.Surface(self.rect.size)
        self.visible = False

        self.onClick = Event()
        self.onClose = Event()

        self.elements = [
            # A blank rectangle.
            SizedPicture(app, self.background, area, area.size),

            # The text.
            TextElement(app, message, font, textPos, textColour,
                anchor=textAnchor
            ),

            # A button to dismiss it.
            TextButton(app,
                pos=buttonPos,
                text='[close]',
                font=font,
                stdColour=closeColour,
                hvrColour=hoverColour,
                onClick=self._doHide
            )
        ]

    #####################
    # Main interface
    #####################

    def show(self, message=None):
        if message is not None:
            self.elements[1].setText(message)
        self._showMessage()

    def hide(self):
        self._hideMessage()

    #####################
    # Element behaviour
    #####################

    def processEvent(self, event):
        '''Processes the specified event and returns the event if it should
        be passed on, or None if it has been caught.'''
        if not self.visible:
            return event

        event = CompoundElement.processEvent(self, event)
        if (event is not None and event.type == pygame.MOUSEBUTTONDOWN and
                self.rect.collidepoint(event.pos)):
            # User has clicked somewhere on the component.
            self.onClick()
            if self.url is not None:
                browser.openPage(self.app, self.url)
            return

        return event

    def draw(self, screen):
        if not self.visible:
            return
        CompoundElement.draw(self, screen)

    def _showMessage(self):
        self.visible = True
        self.area.apply(self.app, self.rect)

        # Show the components.
        self.background = pygame.Surface(self.rect.size)
        self.background.fill(self.backColour)
        self.elements[0].setImage(self.background)

    def _hideMessage(self):
        self.visible = False
        self.onClose()

    def _doHide(self, component):
        self.hide()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

NeutralTeamId = '\x00'
NoPlayer = '\x00'



from math import pi
import random
from trosnoth.ai import AI

from trosnoth.model.upgrades import Grenade
from twisted.internet import reactor, task


class JohnAI(AI):
    '''
    In version 1.1.3 the computer(AI) players have the following functions:
    1. The robot runs "intelligently" around the map until it gets confused
        with a block.
    2. The robot's shooting is randomly off to make it more human like.
    3. The  robot gets confused less
    4. The robot talks less
    5. The robot gets confused by Minimap Disruption
    6. The robot don't shoot players that are invisible.
    7. Uses Grenades when there are 3 players in its current zone
    This file was edited by John "AIMaker" Board at home. 28/12/10 3:57 PM
    '''

    nick = 'JohnBot'
    playable = True

    def start(self):
        self._pauseTimer = None
        self._loop = task.LoopingCall(self.tick)
        self._loop.start(0.5)

    def disable(self):
        self._loop.stop()
        if self._pauseTimer is not None:
            self._pauseTimer.cancel()

    def tick(self):
        self.beforeX = self.player.pos[0]
        if self.player.dead:
            if self.player.inRespawnableZone():
                self.doAimAt(0, thrust=0)
                if self.player.respawnGauge == 0:
                    self.doRespawn()
            else:
                self.aimAtFriendlyZone()
        else:
            if self._pauseTimer is None:
                self.startPlayerMoving()
            if self.player.canShoot():
                self.fireAtNearestEnemy()
            self.useUpgrade()

    def checkEnemyUpgrade(self):
        enemies = [p for p in self.world.players if not
                   (p.dead or self.player.isFriendsWith(p))]
        upgrade = [str(p.upgrade) for p in enemies if p.upgrade is not None]
        return upgrade

    def useUpgrade(self):
        if len(self.getEnemyPlayersInZone()) > 3:
            self.doBuyUpgrade(Grenade)

    def aimAtFriendlyZone(self):
        zones = [z for z in self.world.zones if z.owner == self.player.team]
        if len(zones) == 0:
            return

        def getZoneDistance(zone):
            x1, y1 = self.player.pos
            x2, y2 = zone.defn.pos
            return ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5

        bestZone = min(zones, key=getZoneDistance)
        self.doAimAtPoint(bestZone.defn.pos)

    def fireAtNearestEnemy(self):
        enemies = [
            p for p in self.world.players
            if not (p.dead or p.invisible or self.player.isFriendsWith(p))]
        if len(enemies) == 0:
            return

        def getPlayerDistance(p):
            x1, y1 = self.player.pos
            x2, y2 = p.pos
            return ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5

        nearestEnemy = min(enemies, key=getPlayerDistance)
        # If you get closer than these many pixels you will get shot at, orig
        # value = 512, it is determined by the difficulty rating
        if getPlayerDistance(nearestEnemy) < 300:
            # Do not shoot at ninjas
            if not nearestEnemy.invisible:
                # Add "human" error to nearest enemy pos
                pos = (
                    nearestEnemy.pos[0] + (random.random() * 100),
                    nearestEnemy.pos[1] + (random.random() * 100))
                self.doAimAtPoint(pos)
                self.doShoot()

    def totalOfEnemies(self):
        enemies = [
            p for p in self.world.players
            if not (p.dead or self.player.isFriendsWith(p))]
        return len(enemies)

    def getEnemyPlayersInZone(self):
        return [
            p for p in self.player.getZone().players
            if not (p.dead or self.player.isFriendsWith(p))]

    def getEnemyPlayerCount(self):
        return len(self.getEnemyPlayersInZone())

    def died(self, killer, deathType):
        self._pauseTimer = None
        self.doStop()
        self.aimAtFriendlyZone()

    def respawned(self):
        self.startPlayerMoving()

    def startPlayerMoving(self):
        self.pauseAndReconsider()

    def posOfOrb(self):
        return self.player.getZone().defn.pos

    def pauseAndReconsider(self):
        self.beforeX = self.player.pos[0]
        if self.player.dead:
            self._pauseTimer = None
            return

        # Pause again in between 0.5 and 2.5 seconds time.
        t = random.random() * 0.2 + 0.2
        self._pauseTimer = reactor.callLater(t, self.pauseAndReconsider)

        # Decide on a direction.
        def getPlayerDistance(p):
            x1, y1 = self.player.pos
            x2, y2 = p.pos
            return ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5
        enemies = [
            p for p in self.world.players
            if not (p.dead or self.player.isFriendsWith(p))]
        if len(enemies) > 0:
            nearestEnemy = min(enemies, key=getPlayerDistance)
        if "Minimap Disruption" in self.checkEnemyUpgrade():
                self.randomMove()
        if len(enemies) > 0:
            if self.player.getZone().owner == self.player.team:
                x = nearestEnemy.pos[0]
                y = nearestEnemy.pos[1]
                self.moveToward(x, y)
        elif not self.player.getZone():
            pass
        elif not self.player.getZone().owner == self.player.team:
            if self.getEnemyPlayerCount() == 0:
                x = self.posOfOrb()[0]
                y = self.posOfOrb()[1]
                self.moveToward(x, y)

    def randomMove(self):
        d = random.choice(['drop', 'jump', 'left', 'right'])
        if d == 'drop':
            self.doDrop()
        if d == 'jump':
            self.doJump()
        if d == 'left':
            self.doMoveLeft()
        if d == 'right':
            self.doMoveRight()

    def moveToward(self, x, y):
        if x > self.player.pos[0] and self.player.isAttachedToWall():
            self.doAimAt(pi / 2.)
            self.doMoveRight()
            self.doJump()
        if x < self.player.pos[0] and self.player.isAttachedToWall():
            self.doAimAt(-pi / 2.)
            self.doMoveLeft()
            self.doJump()
        if x > self.player.pos[0] and self.player.isOnGround():
            self.doAimAt(pi / 2.)
            self.doMoveRight()
            self.doJump()
        if x < self.player.pos[0] and self.player.isOnGround():
            self.doAimAt(-pi / 2.)
            self.doMoveLeft()
            self.doJump()
        if y > self.player.pos[1] and self.player.isOnPlatform():
            self.doDrop()
        if y < self.player.pos[1]:
            self.doJump()
        if x == self.player.pos[0] and y < self.player.pos[1]:
            self.doJump()
        if x == self.player.pos[0] and y > self.player.pos[1]:
            self.doDrop()

AIClass = JohnAI



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pygame

from trosnoth.gui.keyboard import VirtualKeySet

# Define virtual keys and their default bindings.
default_game_keys = VirtualKeySet((
    # Movement keys.
    ('left', pygame.K_a),
    ('down', pygame.K_s),
    ('right', pygame.K_d),
    ('jump', pygame.K_w),

    # Used in replay mode.
    ('follow', pygame.K_EQUALS),

    # Menu keys.
    ('menu', pygame.K_ESCAPE),

    ('select upgrade', pygame.K_b),
    ('activate upgrade', pygame.K_SPACE),
    ('change nickname', pygame.K_F12),
    ('more actions', pygame.K_v),
    ('respawn', pygame.K_r),
    ('ready', pygame.K_y),

    ('no upgrade', pygame.K_0),

    ('abandon', pygame.K_m),
    ('chat', pygame.K_RETURN),
    ('leaderboard', pygame.K_p),
    ('toggle interface', pygame.K_DELETE),

    ('toggle terminal', pygame.K_SCROLLOCK),
))

from trosnoth.model.upgrades import allUpgrades

for upgradeClass in allUpgrades:
    if upgradeClass.defaultKey is not None:
        default_game_keys[upgradeClass.action] = upgradeClass.defaultKey
del upgradeClass



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pygame

from trosnoth.gui.framework import framework, elements
from trosnoth.gui.common import Location, translateEvent
from trosnoth.gui import keyboard

class MenuDisplay(framework.CompoundElement):
    '''
    Represents the UI side of a menu. To use this kind of element, you should
    first create a trosnoth.ui.menu.menu.MenuManager object with appropriate
    menus.

    Note that a menu manager will show the hotkeys of actions, but will not
    actually cause the actions to be executed. For this, you should construct
    a gui.hotkey.Hotkeys() object.

    For example:

    menuMan = MenuManager()
    menu1 = Menu(name='file',
                 items=[MenuItem('action1', ...),
                        MenuItem('action2', ...),
                        MenuItem('cancel', menuMan.cancel)])
    menu2 = Menu(name='main',
                 items=[MenuItem('file', lambda: menuMan.show(menu1)),
                        ...])
    menuMan.setDefaultMenu(menu2)

    menuDisp = MenuDisplay(..., menuMan, ...)

    @param  location    should always be an instance of
                        trosnoth.ui.common.Location, but the current
                        implementation accounts for the possibility of it being
                        a point for the top-left.
    '''
    ACCELERATION = 500  # pix/s/s

    def __init__(self, app, location, size, font, manager, titleColour,
                 stdColour, hvrColour, disabledColour=None, backColour=None,
                 autosize=False, hidable=False, keymapping=None):
        super(MenuDisplay, self).__init__(app)

        self.screenSize = app.screenManager.size
        self.location = location
        self.size = size
        self.rect = pygame.Rect(0, 0, 0, 0)
        self.font = font
        self.manager = manager

        self.titleColour = titleColour
        self.stdColour = stdColour
        self.hvrColour = hvrColour
        self.disabledColour = disabledColour or stdColour

        self.autosize = autosize
        self.hidable = hidable
        self.hidden = False
        self.image = None

        self.keymapping = keymapping

        if backColour is None:
            self.backColour = (0, 0, 1)
            self.transparent = True
        else:
            self.backColour = backColour
            self.transparent = False

        if hasattr(size, 'getSize'):
            size = size.getSize(app)
        if not autosize:
            self.image = pygame.Surface(size)
            if self.transparent:
                self.image.set_colorkey(self.backColour)
        self.rect.size = size
        self.velocity = 0

        # When the menu changes we need to be told.
        manager.onShowMenu.addListener(self.refresh)
        self.refresh(manager.menu)

    def _refreshRect(self):
        if hasattr(self.location, 'apply'):
            self.location.apply(self.app, self.rect)
        else:
            self.rect.topleft = self.location

    def drawChildren(self):
        self.image.fill(self.backColour)
        super(MenuDisplay, self).draw(self.image)

    def resizeIfNeeded(self):
        if self.app.screenManager.size == self.screenSize:
            return
        self.screenSize = self.app.screenManager.size

        size = self.size
        if hasattr(size, 'getSize'):
            size = size.getSize(self.app)
        if not self.autosize:
            self.image = pygame.Surface(size)
            if self.transparent:
                self.image.set_colorkey(self.backColour)
        self.rect.size = size
        self.refresh(self.manager.menu)

    def tick(self, deltaT):
        self.resizeIfNeeded()
        diff = self.image.get_rect().height - self.rect.height
        if diff != 0:
            sgn = diff/abs(diff)
            diff = abs(diff)
            self.velocity = sgn * min(abs(self.velocity) +
                                         deltaT*self.ACCELERATION,
                                      (2.0 * self.ACCELERATION * diff)**0.5)
            self.rect.height += sgn * min(abs(self.velocity)*deltaT+1, diff)

        self._refreshRect()
        super(MenuDisplay, self).tick(deltaT)

    def draw(self, screen):
        self.drawChildren()
        if not self.transparent:
            screen.fill(self.backColour, self.rect)
        if self.rect.height < self.image.get_rect().height:
            screen.blit(self.image,
                        self.rect,
                        pygame.Rect((0, 0), self.rect.size))
        else:
            screen.blit(self.image, self.rect)

    def hide(self, btn=None):
        if self.hidden:
            self.hidden = False
            if self.autosize:
                itemSpacing = self.font.getLineSize(self.app)
                size = (self.rect.width,
                        itemSpacing * (1 + len(self.manager.menu.items)))
            else:
                size = self.size
            self.image = pygame.Surface(size)
        else:
            self.hidden = True
            itemSpacing = self.font.getLineSize(self.app)
            self.image = pygame.Surface((self.rect.width, itemSpacing))

        if self.transparent:
            self.image.set_colorkey(self.backColour)
        self.refresh(self.manager.menu)

    def refresh(self, menu=None):
        '''Updates the display of this menu to reflect the currently-showing
        menu of the manager.'''

        if menu is None:
            menu = self.manager.menu
        if menu is None:
            menuItemCount = 0
        else:
            menuItemCount = len(menu.items)

        if self.hidable and self.hidden:
            # Display the menu title only.
            itemSpacing = self.font.getLineSize(self.app)
        elif self.autosize:
            # Make a canvas for the new menu.
            itemSpacing = self.font.getLineSize(self.app)
            size = (self.rect.width, itemSpacing * (1 + menuItemCount))
            self.image = pygame.Surface(size)
            if self.transparent:
                self.image.set_colorkey(self.backColour)

        rect = self.image.get_rect()

        if not (self.autosize or (self.hidable and self.hidden)):
            # Calculate the item spacing.
            itemSpacing = rect.height / (1 + menuItemCount)

        self.elements = []

        if menu is None:
            return

        x = rect.width // 2

        # Menu title.
        if self.hidable:
            stdImage = pygame.Surface((self.rect.width, itemSpacing))
            stdImage.fill(self.backColour)
            img = self.font.render(self.app, menu.name, True, self.titleColour)
            r = img.get_rect()
            r.center = (self.rect.width // 2, itemSpacing // 2)
            stdImage.blit(img, r)

            hvrImage = pygame.Surface((self.rect.width, itemSpacing))
            hvrImage.fill(self.backColour)
            img = self.font.render(self.app, menu.name, True, self.hvrColour)
            hvrImage.blit(img, r)

            btn = elements.HoverButton(app=self.app,
                                     pos=Location((x, itemSpacing/2), 'center'),
                                     stdImage=stdImage,
                                     hvrImage=hvrImage)
            btn.onClick.addListener(self.hide)
            self.elements.append(btn)
            if self.hidden:
                return
        else:
            self.elements.append(
                elements.TextElement(app=self.app, text=menu.name,
                                     font=self.font,
                                     pos=Location((x, itemSpacing / 2),
                                                  'center'),
                                     colour=self.titleColour))

        # Menu items.
        y = itemSpacing * 3 / 2
        for item in menu.items:
            self.addMenuItem(item, x, y, rect.width, itemSpacing)
            y += itemSpacing

    def processEvent(self, event):
        event0 = event
        if event.type in (pygame.MOUSEMOTION, pygame.MOUSEBUTTONDOWN,
                pygame.MOUSEBUTTONUP):
            event = translateEvent(event, self.rect.topleft)
        event1 = super(MenuDisplay, self).processEvent(event)

        # This bit is just so that our event translations are restricted to
        # this menu.

        if event1 is event:
            return event0
        return event1

    def addMenuItem(self, item, x, y, width, itemHeight):
        stdImage = pygame.Surface((self.rect.width, itemHeight))
        hvrImage = pygame.Surface((self.rect.width, itemHeight))
        stdImage.set_colorkey((0,0,0))
        hvrImage.set_colorkey((0,0,0))

        if item.enabled:
            stdColour = self.stdColour
            hvrColour = self.hvrColour
        else:
            stdColour = hvrColour = self.disabledColour

        if item.name == '---':
            # Separator.
            pygame.draw.line(stdImage, stdColour, (0, itemHeight//2),
                             (self.rect.width, itemHeight//2))
            pygame.draw.line(hvrImage, hvrColour, (0, itemHeight//2),
                             (self.rect.width, itemHeight//2))
        else:
            # Put caption on standard canvas.
            img = self.font.render(self.app, item.name, True, stdColour)
            r = img.get_rect()
            r.midleft = (2, itemHeight // 2)
            stdImage.blit(img, r)

            # Put caption on hover canvas.
            img = self.font.render(self.app, item.name, True, hvrColour)
            hvrImage.blit(img, r)

        if item.action is not None and self.keymapping is not None:
            try:
                hkStr = keyboard.shortcutName(self.keymapping.getkey(
                        item.action))
            except KeyError:
                pass
            else:

                # Put hotkey on standard canvas.
                img = self.font.render(self.app, hkStr, True, stdColour)
                r = img.get_rect()
                r.midright = (width-2, itemHeight // 2)
                stdImage.blit(img, r)

                # Put hotkey on hover canvas.
                img = self.font.render(self.app, hkStr, True, hvrColour)
                hvrImage.blit(img, r)

        btn = elements.HoverButton(self.app, Location((x, y), 'center'),
                                   stdImage, hvrImage)
        if item.enabled:
            btn.onClick.addListener(lambda btn: item.execute())
        self.elements.append(btn)



import logging

import pygame

from trosnoth.gui.framework.framework import CompoundElement
from trosnoth.gui.framework.elements import TextElement, TextButton
from trosnoth.gui.framework.prompt import InputBox
from trosnoth.gui.fonts.font import Font
from trosnoth.gui import colours
from trosnoth.gui.common import Location, AttachedPoint, addPositions, Area
from trosnoth.utils.event import Event
from trosnoth.utils.utils import BasicContextManager

log = logging.getLogger('game')


# Styles work thusly:
# When looking for a particular attribute (say backColour),
# a cell will search first in its own style object, then in
# its row's style object, then it its column's style object,
# then in the table's. As soon as it finds a non-None value,
# it will use that.
class TableStyle(BasicContextManager):
    def __init__(self):
        self.backColour = None
        self.foreColour = None
        self.hoverColour = None
        self.hasShadow = None
        self.shadowColour = None
        self.font = None
        self.padding = None
        self.textAlign = None


class StyleUser(CompoundElement):
    def __init__(self, app, style):
        super(StyleUser, self).__init__(app)
        self.style = style


class CellAggregator(StyleUser):
    def __init__(self, app, table):
        super(CellAggregator, self).__init__(app, TableStyle())
        self._table = table
        self._cells = []
        self.index = -1
        self.elements = self._cells

    def setIndex(self, index):
        self.index = index

    def _newCell(self, cell, index=None):
        if index is None:
            index = len(self._cells)
        assert index <= len(self._cells)
        self._cells.insert(index, cell)

    def _anotherInsertedBefore(self):
        self.index += 1

    def _anotherDeletedBefore(self):
        self.index -= 1
        assert self.index >= 0

    def _delCell(self, index):
        del self._cells[index]

    def __len__(self):
        return len(self._cells)

    def __getitem__(self, i):
        return self._cells[i]

    def styleUpdate(self):
        '''
        Informs the row or column that the style of some part of the table has
        changed.  This is provided because it is too expensive for every cell
        to check its style every draw() call, so you must manually tell the
        table, row, column or cell that the style has been updated before the
        update is reflected.
        '''
        for cell in self._cells:
            cell.styleUpdate()


class Row(CellAggregator):
    def __init__(self, app, table, index, height):
        super(Row, self).__init__(app, table)
        assert hasattr(height, 'getVal')
        self.setHeight(height)
        self.setIndex(index)

    def setHeight(self, height):
        assert hasattr(height, 'getVal')
        self._height = height

    def _getHeight(self):
        return self._height.getVal(self.app)

    def _getPt(self):
        return self._table._getRowPt(self.index)


class Column(CellAggregator):
    def __init__(self, app, table, width=0):
        super(Column, self).__init__(app, table)
        self.setWidth(width)

    def onClicked(self, sender):
        self.setFocus(sender)
        self._table.onClicked(self)

    def lostFocus(self):
        self.setFocus(None)

    def setWidth(self, width):
        self._width = width

    def setDefaultWidth(self, width):
        if self._width == 0:
            self._width = width

    def _getWidth(self):
        if hasattr(self._width, 'getVal'):
            return self._width.getVal(self.app)
        else:
            return self._width

    def _getPt(self):
        return self._table._getColPt(self.index)


class TextColumn(Column):
    def createCell(self, row):
        return TextCell(self.app, row, self)


class TextBoxColumn(Column):
    def createCell(self, row):
        return TextBoxCell(self.app, row, self)


class TextButtonColumn(Column):
    def createCell(self, row):
        return TextButtonCell(self.app, row, self)


class Cell(StyleUser):
    def __init__(self, app, row, column):
        super(Cell, self).__init__(app, TableStyle())
        self._row = row
        self._column = column
        self._styleChanged = True
        self.rect = None
        self.usableRect = None

    def layoutChanged(self):
        self.rect = None
        self.usableRect = None

    @property
    def rowNum(self):
        return self._row.index

    @property
    def columnNum(self):
        return self._column.index

    def styleUpdate(self):
        '''
        Informs the cell that the style of some part of the table has changed.
        This is provided because it is too expensive for every cell to check
        its style every draw() call, so you must manually tell the table, row,
        column or cell that the style has been updated before the update is
        reflected.
        '''
        self._styleChanged = True

    def styleGet(self, field):
        toSearch = [self, self._row, self._column, self._row._table]
        for s in toSearch:
            val = getattr(s.style, field)
            if val is not None:
                return val

        return None

    def _getPt(self):
        rowPt = self._row._getPt()
        colPt = self._column._getPt()
        return (colPt[0], rowPt[1])

    def _getSize(self):
        rowHeight = self._row._getHeight()
        colWidth = self._column._getWidth()
        return (colWidth, rowHeight)

    def _getRect(self):
        if self.rect is None:
            self.rect = pygame.Rect(self._getPt(), self._getSize())
        return self.rect

    def _getUsableRect(self):
        if self.usableRect is not None:
            return self.usableRect
        pt = self._getPt()
        padding = self.styleGet('padding')
        usablePt = addPositions(pt, padding)
        size = self._getSize()
        usableSize = tuple([size[i] - padding[i] * 2 for i in 0, 1])
        self.usableRect = pygame.Rect(usablePt, usableSize)
        return self.usableRect

    def _boxClicked(self, sender):
        self.setFocus(sender)
        self._column.onClicked(self)

    def draw(self, surface):
        if self._styleChanged:
            self._styleChanged = False
            self._update()
        bgColour = self.styleGet('backColour')
        if bgColour is not None:
            surface.fill(bgColour, self._getRect())

        clipRect = surface.get_clip()
        newClipRect = clipRect.clip(self._getUsableRect())
        surface.set_clip(newClipRect)
        try:
            super(Cell, self).draw(surface)
        finally:
            surface.set_clip(clipRect)


class TextCell(Cell):
    def __init__(self, app, row, column):
        super(TextCell, self).__init__(app, row, column)
        textAlign = self.styleGet('textAlign')
        self.textElement = TextElement(
            self.app, '', self.styleGet('font'),
            Location(CellAttachedPoint((0, 0), self, textAlign), textAlign),
            colour=self.styleGet('foreColour'))
        self.elements = [self.textElement]
        self._oldText = ''

    def setText(self, text):
        if text != self._oldText:
            self.textElement.setText(text)
            self._styleChanged = True
            self._oldText = text

    def _update(self):
        self.textElement.setFont(self.styleGet('font'))
        self.textElement.setColour(self.styleGet('foreColour'))
        textAlign = self.styleGet('textAlign')
        self.textElement.pos.anchor = textAlign
        self.textElement.pos.point.attachedAt = textAlign
        self.textElement.setShadow(self.styleGet('hasShadow'))
        self.textElement.setShadowColour(self.styleGet('shadowColour'))


class TextBoxCell(Cell):
    def __init__(self, app, row, column):
        super(TextBoxCell, self).__init__(app, row, column)
        textAlign = self.styleGet('textAlign')
        self.inputBox = InputBox(
            self.app,
            Area(
                CellAttachedPoint((0, 0), self, textAlign),
                self._getUsableRect().size, textAlign),
            font=self.styleGet('font'),
            colour=self.styleGet('foreColour'))
        self.inputBox.onEdit.addListener(self._valueChanged)
        self.elements = [self.inputBox]
        self._oldText = ''
        self._readOnly = True
        self.setReadOnly(False)
        self.onValueChanged = Event()

    def _valueChanged(self, sender):
        self._oldText = self.inputBox.getValue()
        self.onValueChanged.execute(self)

    def setReadOnly(self, readOnly):
        if readOnly == self._readOnly:
            return
        self._readOnly = readOnly

        if readOnly:
            self.inputBox.onClick.removeListener(self._boxClicked)
        else:
            self.inputBox.onClick.addListener(self._boxClicked)
            self.setFocus(None)

    def lostFocus(self):
        self.setFocus(None)

    def setText(self, text):
        if text != self._oldText:
            self.inputBox.setValue(text)
            self._styleChanged = True
            self._oldText = text

    def getText(self):
        return self._oldText

    def setMaxLength(self, length):
        self.inputBox.setMaxLength(length)

    def setValidator(self, validator):
        self.inputBox.setValidator(validator)

    def _update(self):
        self.inputBox.setFont(self.styleGet('font'))
        self.inputBox.setFontColour(self.styleGet('foreColour'))
        bgColour = self.styleGet('backColour')
        if bgColour is not None:
            self.inputBox.setBackColour(bgColour)
        textAlign = self.styleGet('textAlign')
        self.inputBox.area.anchor = textAlign
        self.inputBox.area.point.attachedAt = textAlign


class TextButtonCell(Cell):
    def __init__(self, app, row, column):
        super(TextButtonCell, self).__init__(app, row, column)
        textAlign = self.styleGet('textAlign')
        pos = Location(CellAttachedPoint((0, 0), self, textAlign), textAlign)
        self.textButton = TextButton(
            self.app, pos, '', self.styleGet('font'),
            self.styleGet('foreColour'), self.styleGet('hoverColour'))
        self.elements = [self.textButton]
        self._oldText = ''

    def setOnClick(self, onClick):
        self.textButton.onClick.addListener(onClick)
        self._boxClicked(None)

    def setText(self, text):
        if text != self._oldText:
            self.textButton.setText(text)
            self._styleChanged = True
            self._oldText = text

    def _update(self):
        self.textButton.setFont(self.styleGet('font'))
        self.textButton.setColour(self.styleGet('foreColour'))
        self.textButton.setHoverColour(self.styleGet('hoverColour'))
        self.textButton.setBackColour(self.styleGet('backColour'))
        textAlign = self.styleGet('textAlign')
        self.textButton.pos.anchor = textAlign
        self.textButton.pos.point.attachedAt = textAlign


class CellAttachedPoint(AttachedPoint):
    def __init__(self, val, cell, attachedAt='topleft'):
        super(CellAttachedPoint, self).__init__(
            val, cell._getUsableRect, attachedAt)


class Value(object):
    def __init__(self, value):
        self._value = value

    def getVal(self):
        return self._value


class Table(StyleUser):
    def __init__(self, app, pos, rows=None, columns=None, topPadding=0):
        defaultStyle = TableStyle()
        defaultStyle.backColour = (255, 255, 255)
        defaultStyle.padding = (5, 5)
        defaultStyle.foreColour = (0, 0, 0)
        defaultStyle.hoverColour = (0, 0, 0)
        defaultStyle.font = Font('FreeSans.ttf', 36)
        defaultStyle.textAlign = 'topleft'
        defaultStyle.hasShadow = False
        defaultStyle.shadowColour = colours.shadowDefault
        super(Table, self).__init__(app, defaultStyle)
        self.pos = pos
        self.topPadding = topPadding

        self._columns = []
        self._rows = []
        self.elements = self._columns

        # Style Settings:
        self._borderColour = (192, 192, 192)
        self._borderWidth = 10
        self._defaultWidth = 200
        self.setDefaultHeight(Value(60))

        if rows is not None:
            self.addRows(rows)

        if columns is not None:
            self.addEmptyColumns(columns)

        self._sizes = {}
        self._appSize = None
        app.screenManager.onResize.addListener(self.appResized)

    def appResized(self):
        for row in self._rows:
            for cell in row._cells:
                cell.layoutChanged()

    def onClicked(self, sender):
        self.setFocus(sender)

    def setDefaultHeight(self, height):
        self._defaultHeight = height

    def styleUpdate(self):
        '''
        Informs the table that the style of some part of the table has changed.
        This is provided because it is too expensive for every cell to check
        its style every draw() call, so you must manually tell the table, row,
        column or cell that the style has been updated before the update is
        reflected.
        '''
        for row in self._rows:
            row.styleUpdate()

    def setBorderWidth(self, width):
        if width != self._borderWidth:
            self._borderWidth = width
            self._sizes = {}

    def setBorderColour(self, colour):
        self._borderColour = colour

    def addRow(self, index=None):
        if index is None:
            index = len(self._rows)
        assert index <= len(self._rows)
        # Create and Insert
        newRow = Row(self.app, self, index, self._defaultHeight)
        self._rows.insert(index, newRow)
        for x in range(0, len(self._columns)):
            newCell = self._columns[x].createCell(newRow)
            newRow._newCell(newCell, x)
            self._columns[x]._newCell(newCell, index)

        for x in range(index+1, len(self._rows)):
            self._rows[x]._anotherInsertedBefore()
        self._sizes = {}
        return newRow

    def addColumn(self, column, index=None):
        if index is None:
            index = len(self._columns)
        if index > len(self._columns):
            raise ValueError('invalid index %r' % (index,))

        column.setIndex(index)
        column.setDefaultWidth(self._defaultWidth)

        self._columns.insert(index, column)
        for rowIndex, row in enumerate(self._rows):
            newCell = Cell(self.app, row, column)
            column._newCell(newCell, rowIndex)
            row._newCell(newCell, index)

        for column in self._columns[index+1:]:
            column._anotherInsertedBefore()
        self._sizes = {}

    def addColumns(self, columns):
        for column in columns:
            self.addColumn(column)

    def addRows(self, count, firstIndex=None):
        if count < 1:
            return

        index = firstIndex

        for i in range(count):
            self.addRow(index)
            if index is not None:
                index += 1

    def addEmptyColumns(self, count, firstIndex=None):
        if count < 1:
            return

        index = firstIndex

        for i in range(count):
            self.addColumn(
                TextColumn(self.app, self, self._defaultWidth), index)
            if index is not None:
                index += 1

    def clear(self):
        while self._rows:
            self.delRow(self.rowCount() - 1)

    def delRow(self, index):
        del self._rows[index]
        for col in self._columns:
            col._delCell(index)
        for x in range(index, len(self._rows)):
            self._rows[x]._anotherDeletedBefore()
        self._sizes = {}

    def delColumn(self, index):
        del self._columns[index]
        for row in self._rows:
            row._delCell(index)
        for x in range(index, len(self._columns)):
            self._columns[x]._anotherDeletedBefore()
        self._sizes = {}

    def rowCount(self):
        return len(self._rows)

    def columnCount(self):
        return len(self._columns)

    def getRect(self, app):
        assert app == self.app
        return self._getRect()

    def _getRect(self):
        r = pygame.Rect((0, 0), self._getSize())
        self.pos.apply(self.app, r)
        return r

    def _getPt(self):
        x, y = self._getRect().topleft
        y += self.topPadding
        return x, y

    def _getSize(self, rowStart=0, rowEnd=None, colStart=0, colEnd=None):
        if rowEnd is None:
            rowEnd = len(self._rows)
        if colEnd is None:
            colEnd = len(self._columns)

        appSize = self.app.screenManager.size
        if appSize != self._appSize:
            self._sizes = {}
            self._appSize = appSize
        else:
            try:
                return self._sizes[rowStart, rowEnd, colStart, colEnd]
            except KeyError:
                pass

        y = self._borderWidth
        for j in range(rowStart, rowEnd):
            y += self._rows[j]._getHeight()
            y += self._borderWidth

        x = self._borderWidth
        for i in range(colStart, colEnd):
            x += self._columns[i]._getWidth()
            x += self._borderWidth

        self._sizes[rowStart, rowEnd, colStart, colEnd] = (x, y)
        return (x, y)

    def _getRowPt(self, index):
        pt = self._getPt()
        offset = self._getSize(0, index)
        return (pt[0], pt[1]+offset[1])

    def _getColPt(self, index):
        pt = self._getPt()
        offset = self._getSize(colStart=0, colEnd=index)
        return (pt[0]+offset[0], pt[1])

    def draw(self, surface):
        super(Table, self).draw(surface)
        if self._borderColour is not None:
            self._drawBorders(surface)

    def _getBorderWidth(self):
        if hasattr(self._borderWidth, 'getVal'):
            return self._borderWidth.getVal(self.app)
        else:
            return self._borderWidth

    def _drawBorders(self, surface):
        width = self._getBorderWidth()
        if width == 0:
            return
        if self.rowCount() == 0 or self.columnCount() == 0:
            return
        # Draw Horizontals:
        # Use rect filling, because drawing lines usually round the ends
        horRect = pygame.Rect(
            (0, 0), (self._getSize(rowStart=0, rowEnd=1)[0], width))
        for row in self._rows:
            horRect.bottomleft = row._getPt()
            surface.fill(self._borderColour, horRect)
        horRect.top += row._getHeight() + width
        surface.fill(self._borderColour, horRect)

        # Draw Verticals:
        # Use rect filling, because drawing lines usually round the ends
        verRect = pygame.Rect(
            (0, 0), (width, self._getSize(colStart=0, colEnd=1)[1]))
        for col in self._columns:
            verRect.topright = col._getPt()
            surface.fill(self._borderColour, verRect)
        verRect.left += col._getWidth() + width
        surface.fill(self._borderColour, verRect)

    def __getitem__(self, index):
        return self.getColumn(index)

    def getRow(self, index):
        return self._rows[index]

    def getColumn(self, index):
        return self._columns[index]
